from ._shared import *

class NavigatorIOMixin:
    def infer_axes_from_shape(shape):
        """
        Build an ImageJ-style axes string from a NumPy shape tuple:
        - T (time)   for the first dim if >1
        - Z (z-slice) for the next if >1
        - C (channel) for the next if >1
        - Y, X always the last two dims
        """
        axes = []
        letter_map = ['T', 'Z', 'C']
        for i, length in enumerate(shape[:-2]):
            axes.append(letter_map[i] if length > 1 else letter_map[i].lower())
        axes += ['Y', 'X']
        # Only keep the “real” axes (uppercase)
        return ''.join(ax for ax in axes if ax.isupper())

    def handle_movie_load(self, fname=None, pixelsize=None, frameinterval=None):
        # If save_and_load_routine is active, don't open the dialog.
        if self.save_and_load_routine:
            # Reset the flag so it only applies once.
            self.save_and_load_routine = False
        else:
            # Open the file dialog.
            fname, _ = QFileDialog.getOpenFileName(
                self, "Open Movie TIFF", self._last_dir, "TIFF Files (*.tif *.tiff)"
            )
        
        # If no file was chosen, exit.
        if not fname:
            return

        # Pass the chosen filename to load_movie.
        self.load_movie(fname, pixelsize=pixelsize, frameinterval=frameinterval)

    def load_movie(self, fname=None, pixelsize=None, frameinterval=None):

        self.clear_flag = False
        self.cancel_left_click_sequence()

        self.show_steps = False
        self.showStepsAction.setChecked(False)

        if (self.rois or self.kymographs or
            (hasattr(self, 'trajectoryCanvas') and self.trajectoryCanvas.trajectories)):
            reply = QMessageBox.question(
                self,
                "Clear existing data?",
                "Clear existing data before loading a new movie?",
                QMessageBox.Yes | QMessageBox.Cancel,
                QMessageBox.Yes
            )
            if reply == QMessageBox.Cancel:
                # user chose cancel → abort load_movie entirely
                return
            # user chose Yes → clear all existing data
            self.clear_flag = True

        if fname:
            self._last_dir = os.path.dirname(fname)
            # try:
            with tifffile.TiffFile(fname) as tif:
                temp_movie = tif.asarray()
                self.movie_metadata = tif.imagej_metadata or {}
                page = tif.pages[0]

            tags = page.tags
            description = page.description

            if pixelsize is not None and frameinterval is not None:
                self.pixel_size = pixelsize
                self.frame_interval = frameinterval

            else:
                
                y_size = None
                self.pixel_size = None
                self.frame_interval = None

                import re
                
                if description:
                    # Look for a pattern like "Voxel size: 0.1100x0.1100x1"
                    m = re.search(r'Voxel size:\s*([\d\.]+)[xX]([\d\.]+)[xX]([\d\.]+)', description)
                    if m:
                        try:
                            # Convert the strings to floats.
                            # The typical order in ImageJ is "x_size x y_size x z_size".
                            #x_size = float(m.group(1))
                            y_size = float(m.group(2))
                            #z_size = float(m.group(3))
                            # Return as [z, y, x]
                            self.pixel_size = y_size*1000
                        except Exception as e:
                            print("Error parsing voxel size from ImageDescription:", e)

                if self.pixel_size is None:
                    if 'YResolution' in tags:
                        value = tags['YResolution'].value
                        try:
                            # If value is a tuple, compute pixels per micron:
                            num, denom = value
                            # pixels per micron = num/denom; thus pixel size in microns is 1/(num/denom)
                            self.pixel_size = float(denom)*1000 / float(num)
                        except Exception:
                            try:
                                # Otherwise try to convert directly to a float.
                                self.pixel_size = float(value)*1000
                            except Exception:
                                pass

                desc = tif.pages[0].tags["ImageDescription"].value
                try:
                    match = re.search(r'finterval=([\d\.]+)', desc)
                    if match:
                        self.frame_interval = float(match.group(1))*1000
                except Exception:
                    pass

                if self.pixel_size is not None:
                    if self.pixel_size < 0.1:
                        self.pixel_size = None
                if self.frame_interval is not None:
                    if self.frame_interval < 0.1:
                        self.frame_interval = None

            try:
                shape = temp_movie.shape
            except Exception as e:
                QMessageBox.critical(self, "Error", f"Invalid movie : {e}")
                return

            if temp_movie.ndim not in (3, 4):
                QMessageBox.critical(self, "Error", f"Invalid movie shape: {temp_movie.shape}")
                return

            self.referenceImage = None
            self.refBtn.setVisible(False)
            self.refBtn.setChecked(False)
            self.sumBtn.setChecked(False)
            self.zoomInsetFrame.setVisible(False)
            self.movieCanvas._last_inset_params = None
            self.movieCanvas._inset_update_pending = False

            # Initialize (or reset) contrast settings for multi‑channel movies.
            self.channel_contrast_settings = {}      # for “normal” mode
            self.channel_sum_contrast_settings = {}   # for sum‐mode
            self.reference_contrast_settings = {}

            self.movieNameLabel.setText("")
            self.movieNameLabel.setStyleSheet("background: transparent; color: black; font-size: 16px; font-weight: bold")
            self.movieNameLabel.setText(os.path.basename(fname))
            self.movieNameLabel.setAttribute(Qt.WA_TransparentForMouseEvents, True)
            

            # ── blank out the histogram canvas entirely ────────────────────────────────
            if hasattr(self, 'histogramCanvas'):
                self.histogramCanvas.ax.cla()
                self.histogramCanvas.ax.axis("off")
                self.histogramCanvas.draw_idle()
            # ── blank out the intensity/plot canvas (two sub‐axes) ────────────────────
            if hasattr(self, 'intensityCanvas'):
                ic = self.intensityCanvas

                # 1) clear any old highlights & scatter
                ic.clear_highlight()
                ic.scatter_obj_top = None
                ic.scatter_obj_bottom = None
                ic.point_highlighted = False
                ic._last_plot_args   = None

                # 2) completely wipe both axes
                ic.ax_top.cla()
                ic.ax_bottom.cla()
                ic.ax_top.axis('off')
                ic.ax_bottom.axis('off')

                # 3) redraw & grab a fresh background
                ic.draw()
                ic._background = ic.copy_from_bbox(ic.fig.bbox)
            # ── blank out the velocity histogram canvas ───────────────────────────────
            if hasattr(self, 'velocityCanvas'):
                self.velocityCanvas.ax_vel.cla()
                self.velocityCanvas.ax_vel.axis("off")
                self.velocityCanvas.draw_idle()

            # ── blank out the zoom‐inset widget ─────────────────────────────────────────
            if hasattr(self, "zoomInsetFrame"):
                # hide the whole frame
                self.zoomInsetFrame.setVisible(False)

            if hasattr(self, "zoomInsetWidget"):
                # clear its axes
                self.zoomInsetWidget.ax.clear()
                self.zoomInsetWidget.ax.axis("off")
                self.zoomInsetWidget.draw_idle()

            self.movieCanvas.clear_canvas()
            self.movie = temp_movie
            self.original_movie = self.movie.copy()

            # Reset the frame cache whenever a new movie is loaded.
            self.frame_cache = {}

            if self.movie.ndim == 4:
                # 4D movie (multi‑channel)
                if self.movie_metadata and "axes" in self.movie_metadata:
                    axes_str = self.movie_metadata["axes"]  # e.g., "TXYC" or "TCYX"
                    # For example, if the letter "C" is found, use that index:
                    self._channel_axis = axes_str.find("C")
                else:
                    self.update_channel_axis_options()
                self.update_movie_channel_combo()
                first_frame = self.get_movie_frame(0)
            else:
                # 3D movie (single channel)
                self._channel_axis = None
                self.movieChannelCombo.blockSignals(True)
                self.movieChannelCombo.clear()
                self.movieChannelCombo.addItem("1")
                self.movieChannelCombo.blockSignals(False)
                # Now call update_movie_channel_combo even for 3D movies:
                self.update_movie_channel_combo()
                first_frame = self.movie[0]

            max_frame = self.movie.shape[0]
            self.frameSlider.setMinimum(0)
            self.frameSlider.setMaximum(max_frame - 1)
            self.frameSlider.setValue(0)
            self.frameNumberLabel.setText("1")

            margin = 0
            full_width = self.movieCanvas.image.shape[1]
            full_height = self.movieCanvas.image.shape[0]
            self.movieCanvas.zoom_center = (full_width/2, full_height/2)
            self.movieCanvas.display_image(first_frame)

            if self.clear_flag:
                self.clear_rois()
                self.clear_kymographs()
                self.trajectoryCanvas.clear_trajectories(prompt=False)
                self.clear_flag = False

            self.movieCanvas.draw_idle()
            self.movieCanvas.clear_sum_cache()

            self.update_scale_label()

            if self.pixel_size is None or self.frame_interval is None:
                prev_px = self.pixel_size
                prev_ft = self.frame_interval
                self.set_scale()
                if self.pixel_size is None or self.frame_interval is None:
                    if prev_px is not None or prev_ft is not None:
                        self.pixel_size = None
                        self.frame_interval = None
                        self.update_scale_label()

            # self.last_kymo_by_channel = {}

            # If there are any existing custom columns, ask whether to clear them now
            # only ask if there are any custom‐columns other than the auto‐added colocalization % ones
            d_col = getattr(self, "_DIFF_D_COL", None)
            a_col = getattr(self, "_DIFF_A_COL", None)
            auto_cols = {c for c in (d_col, a_col) if c}
            non_coloc = [
                name for name in self.trajectoryCanvas.custom_columns
                if self.trajectoryCanvas._column_types.get(name) != "coloc"
                and name not in auto_cols
            ]
            if non_coloc:
                reply = QMessageBox.question(
                    self,
                    "Clear custom columns?",
                    "Do you want to clear all user-defined custom columns?",
                    QMessageBox.Yes | QMessageBox.No,
                    QMessageBox.Yes
                )
                if reply == QMessageBox.Yes:
                    # remove only those non-coloc columns
                    for name in non_coloc:
                        idx = self.trajectoryCanvas._col_index.get(name)
                        if idx is not None:
                            self.trajectoryCanvas._remove_custom_column(idx, name)
                    # rebuild the menu so it no longer shows those entries
            
            coloc_cols = [
                name for name in self.trajectoryCanvas.custom_columns
                if self.trajectoryCanvas._column_types.get(name) == "coloc"
            ]
            for name in coloc_cols:
                idx = self.trajectoryCanvas._col_index.get(name)
                if idx is not None:
                    self.trajectoryCanvas._remove_custom_column(idx, name)
            self._rebuild_color_by_actions()

            if self.movie.ndim == 4:
                # — multi-channel: build co-localization columns —
                # first, clear any old co. % columns if reloading
                old = [c for c in self.trajectoryCanvas.custom_columns if c.endswith(" co. %")]
                for name in old:
                    idx = self.trajectoryCanvas._col_index.get(name)
                    if idx is not None:
                        self.trajectoryCanvas._remove_custom_column(idx, name)

                # now add fresh ones
                self.trajectoryCanvas.custom_columns = [c for c in self.trajectoryCanvas.custom_columns if not c.endswith(" co. %")]
                n_chan = self.movie.shape[self._channel_axis]
                for ch in range(1, n_chan+1):
                    col_name = f"Ch. {ch} co. %"
                    self.trajectoryCanvas._add_custom_column(col_name, col_type="coloc")

                # print("load_movie custom_columns, after add_custom_column", self.trajectoryCanvas.custom_columns)

                if not hasattr(self, 'colocalizationAction'):
                    self.colocalizationAction = QAction("Check colocalization", self, checkable=True)
                    self.colocalizationAction.setChecked(False)
                    self.colocalizationAction.toggled.connect(self.on_colocalization_toggled)
                    # spotMenu must be the original QMenu.
                    self.spotMenu.addAction(self.colocalizationAction)

            else:
                # single-channel: ensure no stray co. % columns remain
                for name in [c for c in self.trajectoryCanvas.custom_columns if c.endswith(" co. %")]:
                    idx = self.trajectoryCanvas._col_index.get(name)
                    if idx is not None:
                        self.trajectoryCanvas._remove_custom_column(idx, name)

                self.check_colocalization = False
                # if we previously added the colocalization action, uncheck & remove it
                if hasattr(self, 'colocalizationAction'):
                    # 1) uncheck so UI state is clean
                    self.colocalizationAction.setChecked(False)
                    # 2) disconnect its signal to avoid any stray callbacks
                    try:
                        self.colocalizationAction.toggled.disconnect(self.on_colocalization_toggled)
                    except TypeError:
                        pass
                    # 3) remove it from the spot‐menu
                    self.spotMenu.removeAction(self.colocalizationAction)
                    # 4) delete the attribute so next time we know it’s gone
                    del self.colocalizationAction

            # Always untoggle any active “Color by …” before we potentially reload columns
            self.set_color_by(None)

            self.flash_message("Loaded movie")

            if self.movie.ndim == 4:
                filt = self._ch_overlay._bubble_filter
                filt._wobj = self._ch_overlay
                QTimer.singleShot(2000, lambda: filt._showBubble(force=True))

            # except Exception as e:
            #     QMessageBox.critical(self, "Error", f"Could not load movie:\n{str(e)}")

    def get_movie_frame(self, frame_idx, channel_override=None):
        if self.movie is None:
            return None

        # For 4D movies, include selected channel and channel axis in the cache key.
        if self.movie.ndim == 4:
            if channel_override is None:
                # Use the main GUI's channel selection.
                selected_chan = int(self.movieChannelCombo.currentText()) - 1
            else:
                # Use the override value.
                selected_chan = channel_override-1
            channel_axis = self._channel_axis
            cache_key = (frame_idx, selected_chan, channel_axis)
        else:
            cache_key = frame_idx

        # Initialize cache if it doesn't exist.
        if not hasattr(self, "frame_cache"):
            self.frame_cache = {}

        # Return the frame from the cache if available.
        if cache_key in self.frame_cache:
            return self.frame_cache[cache_key]

        try:
            if self.movie.ndim == 4:
                idx = [0] * self.movie.ndim
                idx[0] = frame_idx
                for ax in range(1, self.movie.ndim):
                    idx[ax] = selected_chan if ax == channel_axis else slice(None)
                frame = self.movie[tuple(idx)]
            else:
                frame = self.movie[frame_idx]
            # Store the computed frame in the cache.
            self.frame_cache[cache_key] = frame
            return frame
        except IndexError:
            print(f"index {frame_idx} out of bounds.")
            return None

    def on_channel_axis_changed(self):
        """
        Called when the user changes the channel axis selection.
        Verify that the selected axis is valid for the loaded movie.
        If the axis is invalid, display an error popup and revert to the previous working axis.
        """
        new_axis = self._channel_axis  # set by ChannelAxisDialog
        old_axis = getattr(self, "_channel_axis", 1)

        # If no movie is loaded, simply update the stored axis.
        if self.movie is None:
            self._channel_axis = new_axis
            return

        try:
            # Try accessing the movie dimension with the new axis.
            _ = self.movie.shape[new_axis]
            # If valid, update the stored channel axis.
            self._channel_axis = new_axis
            # Update the channel dropdown that depends on the channel axis.
            self.update_movie_channel_combo()
        except Exception as e:
            # If the axis is invalid, show an error and revert to the previous value.
            QMessageBox.critical(self, "Error", f"Invalid channel axis: {new_axis}.\nError: {str(e)}")
            self._channel_axis = old_axis

    def update_channel_axis_options(self):
        """
        Update the stored channel axis using the movie's available axes (excluding axis 0).
        Choose as default the axis with the smallest size (typically the channel axis).
        """
        if self.movie is None or self.movie.ndim != 4:
            return

        # List candidate axes: all axes except axis 0.
        candidate_axes = list(range(1, self.movie.ndim))
        # Pick the axis with the smallest size.
        default_axis = min(candidate_axes, key=lambda ax: self.movie.shape[ax])
        self._channel_axis = default_axis

        # Update the channel combo box options.
        self.update_movie_channel_combo()

    def on_channel_changed(self, index):
        # only multi-channel movies
        if self.movie is None or self.movie.ndim != 4:
            return

        if self.looping:
            self.stoploop()

        if self.refBtn.isChecked():
            self.refBtn.setChecked(False)

        self.cancel_left_click_sequence()

        # 1) figure out which channel we’re on
        ch = index + 1

        # 2) refresh kymographs
        self.update_kymo_list_for_channel()

        # 3) pick the right contrast-settings dict
        if self.sumBtn.isChecked():
            settings_store = self.channel_sum_contrast_settings
            display_fn     = self.movieCanvas.display_sum_frame
        else:
            settings_store = self.channel_contrast_settings
            display_fn     = None  # we’ll use update_image_data below

        # 4) if first time for this channel, compute & stash defaults
        if ch not in settings_store:
            # grab the very first frame of this channel
            frame0 = self.get_movie_frame(0)
            p15, p99 = np.percentile(frame0, (15, 99))
            vmin0    = int(p15 * (1.05 if self.sumBtn.isChecked() else 1.0))
            vmax0    = int(p99 * (1.20 if self.sumBtn.isChecked() else 1.10))
            delta    = vmax0 - vmin0
            settings_store[ch] = {
                'vmin':         vmin0,
                'vmax':         vmax0,
                'extended_min': vmin0 - int(0.7 * delta),
                'extended_max': vmax0 + int(1.4 * delta)
            }

        # 5) pull out the stored settings
        s = settings_store[ch]

        # 6) update the slider so it “sticks”
        slider = self.contrastControlsWidget.contrastRangeSlider
        slider.blockSignals(True)
        slider.setMinimum(s['extended_min'])
        slider.setMaximum(s['extended_max'])
        slider.setRangeValues(s['vmin'], s['vmax'])
        slider.blockSignals(False)

        # 7) push the contrast into the MovieCanvas
        mc = self.movieCanvas
        mc._default_vmin = s['vmin']
        mc._default_vmax = s['vmax']
        mc._vmin         = s['vmin']
        mc._vmax         = s['vmax']
        if mc._im is not None:
            mc._im.set_clim(s['vmin'], s['vmax'])

        # 8) redraw either sum‐mode or normal‐mode
        if self.sumBtn.isChecked():
            mc.clear_sum_cache()
            mc.display_sum_frame()
        else:
            # keep the same time‐point
            frame_idx = self.frameSlider.value()
            frame     = self.get_movie_frame(frame_idx)
            mc.update_image_data(frame)

        self._ch_overlay.setText(f"ch{ch}")
        self._ch_overlay.adjustSize()
        self._reposition_channel_overlay()
        self._ch_overlay.show()

        # 10) clear & redraw trajectories on the movie, now that channel has changed
        self.movieCanvas.clear_movie_trajectory_markers()
        self.movieCanvas.draw_trajectories_on_movie()

        if self.intensityCanvas.point_highlighted and ch == self.analysis_channel and self.intensityCanvas._last_plot_args is not None:

            ic_index = self.intensityCanvas.current_index

            # cache arrays once
            centers = np.asarray(self.analysis_search_centers)  # shape (N,2)
            cx, cy = centers[ic_index]

            mc.overlay_rectangle(cx, cy, int(2*self.searchWindowSpin.value()))
            mc.remove_gaussian_circle()

            fc = fs = pk = None
            # draw fit circle & intensity highlight
            if hasattr(self, "analysis_fit_params") and ic_index < len(self.analysis_fit_params):
                fc, fs, pk = self.analysis_fit_params[ic_index]

            pointcolor = self.intensityCanvas.get_current_point_color()
            mc.add_gaussian_circle(fc, fs, pointcolor)

            # only overlay kymo marker if ROI present
            kymo_name = self.kymoCombo.currentText()
            # look up its channel in the map
            info = self.kymo_roi_map.get(kymo_name, {})
            current_kymo_ch = info.get("channel", None)
            if self.analysis_channel == current_kymo_ch or self.analysis_channel is None:
                kymo_name = self.kymoCombo.currentText()
                if kymo_name and kymo_name in self.kymographs and self.rois:
                    roi = self.rois[self.roiCombo.currentText()]
                    xk = None
                    # check fit‐center first, then raw center
                    if fc is not None and is_point_near_roi(fc, roi):
                        xk = self.compute_kymo_x_from_roi(
                            roi, fc[0], fc[1],
                            self.kymographs[kymo_name].shape[1]
                        )
                    elif is_point_near_roi((cx, cy), roi):
                        xk = self.compute_kymo_x_from_roi(
                            roi, cx, cy,
                            self.kymographs[kymo_name].shape[1]
                        )
                    if xk is not None:
                        disp_frame = (self.movie.shape[0] - 1) - self.frameSlider.value()
                        self.kymoCanvas.add_circle(
                            xk, disp_frame,
                            color=pointcolor if fc is not None else 'grey'
                        )


        self.movieCanvas.draw()

    def _on_overlay_clicked(self):
        # build a stand-alone QMenu
        menu = QMenu(None)
        menu.setWindowFlags(Qt.Popup | Qt.FramelessWindowHint)
        menu.setAttribute(Qt.WA_TranslucentBackground)

        # populate it
        n_channels = (self.movie.shape[self._channel_axis]
                    if self.movie is not None and self.movie.ndim == 4 else 1)
        for i in range(n_channels):
            action = menu.addAction(f"ch{i+1}")
            action.setData(i)

        # compute global positions
        lbl = self._ch_overlay
        lbl_global = lbl.mapToGlobal(QPoint(0, 0))
        lbl_width  = lbl.width()
        # measure menu width from its sizeHint
        menu_width = menu.sizeHint().width()

        # center the menu’s x under the label
        x = lbl_global.x() + (lbl_width  - menu_width)//2
        # drop it just below the label
        y = lbl_global.y() + lbl.height()

        chosen = menu.exec_(QPoint(x, y))
        if chosen:
            idx = chosen.data()
            self.movieChannelCombo.setCurrentIndex(idx)

    def _reposition_channel_overlay(self):
        lbl = self._ch_overlay
        # 10px from left, 10px from top
        x = 10
        y = 10
        lbl.move(x, y)

    def update_movie_channel_combo(self, flash=False):
        if self.movie is None:
            self.channelControlContainer.setVisible(False)
            self._ch_overlay.hide()
            return

        current_channel = 1
        if self.movie.ndim == 4:
            self.channelControlContainer.setVisible(False) #OVERRIDE
            channel_axis = self._channel_axis
            try:
                current_channel = int(self.movieChannelCombo.currentText())
            except Exception:
                current_channel = 1

            self._ch_overlay.setText(f"ch{current_channel}")
            self._ch_overlay.adjustSize()
            self._reposition_channel_overlay()
            self._ch_overlay.show()

            self.movieChannelCombo.blockSignals(True)
            self.movieChannelCombo.clear()
            num_channels = self.movie.shape[channel_axis]
            for i in range(num_channels):
                self.movieChannelCombo.addItem(str(i + 1))
            self.movieChannelCombo.setCurrentIndex(current_channel - 1)
            self.movieChannelCombo.blockSignals(False)
            self.movieChannelCombo.setEnabled(True)

            # Get the first frame for the selected channel if needed.
            first_frame = self.get_movie_frame(0)
        else:
            self.channelControlContainer.setVisible(False)
            self._ch_overlay.hide()
            self.movieChannelCombo.blockSignals(True)
            self.movieChannelCombo.clear()
            self.movieChannelCombo.addItem("1")
            self.movieChannelCombo.blockSignals(False)
            self.movieChannelCombo.setEnabled(False)
            first_frame = self.movie[0]
            # Set default current_channel for 3D movies:
            current_channel = 1
    
        # Branch on the current mode (sum vs. normal) and obtain contrast settings.
        if self.sumBtn.isChecked():
            self.movieCanvas.clear_sum_cache()
            self.movieCanvas.display_sum_frame()
            # Sum mode – use channel_sum_contrast_settings.
            if current_channel not in self.channel_sum_contrast_settings:
                p15, p99 = np.percentile(first_frame, (15, 99))
                default_vmin = int(p15 * 1.05)
                default_vmax = int(p99 * 1.2)
                delta = default_vmax - default_vmin
                settings = {
                    'vmin': default_vmin,
                    'vmax': default_vmax,
                    'extended_min': default_vmin - int(0.7 * delta),
                    'extended_max': default_vmax + int(1.4 * delta)
                }
                self.channel_sum_contrast_settings[current_channel] = settings
            else:
                settings = self.channel_sum_contrast_settings[current_channel]
        else:
            # Normal mode – use channel_contrast_settings.
            if current_channel not in self.channel_contrast_settings:
                p15, p99 = np.percentile(first_frame, (15, 99))
                default_vmin = int(p15)
                default_vmax = int(p99 * 1.1)
                delta = default_vmax - default_vmin
                settings = {
                    'vmin': default_vmin,
                    'vmax': default_vmax,
                    'extended_min': default_vmin - int(0.7 * delta),
                    'extended_max': default_vmax + int(1.4 * delta)
                }
                self.channel_contrast_settings[current_channel] = settings
            else:
                settings = self.channel_contrast_settings[current_channel]

        if flash:
            self.flash_message(f"Channel {current_channel}")
        #print(current_channel, settings)

        # Update the movie canvas's internal contrast defaults.
        self.movieCanvas._default_vmin = settings['vmin']
        self.movieCanvas._default_vmax = settings['vmax']
        self.movieCanvas._vmin = settings['vmin']
        self.movieCanvas._vmax = settings['vmax']

        # Update the contrast slider.
        self.contrastControlsWidget.contrastRangeSlider.blockSignals(True)
        self.contrastControlsWidget.contrastRangeSlider.setMinimum(settings['extended_min'])
        self.contrastControlsWidget.contrastRangeSlider.setMaximum(settings['extended_max'])
        self.contrastControlsWidget.contrastRangeSlider.setRangeValues(settings['vmin'], settings['vmax'])
        self.contrastControlsWidget.contrastRangeSlider.blockSignals(False)
        self.contrastControlsWidget.contrastRangeSlider.update()

        # Finally, display the first frame with the correct contrast.
        self.movieCanvas.update_image_data(first_frame)

    def load_reference(self):
        if self.movie is None:
            QMessageBox.warning(self, "", 
                "Please load a movie before loading a reference.")
            return
        fname, _ = QFileDialog.getOpenFileName(
            self, "Open Reference Image", "", "Image Files (*.tif *.tiff *.png *.jpg)"
        )
        if not fname:
            return
        try:
            ref_img = tifffile.imread(fname)
            if ref_img.ndim == 3:
                # Heuristic to decide if the image is multi‐channel:
                small_first = ref_img.shape[0] <= 4 and ref_img.shape[0] > 1
                small_last = ref_img.shape[-1] <= 4 and ref_img.shape[-1] > 1

                if small_first and small_last:
                    choice, ok = QtWidgets.QInputDialog.getItem(
                        self, "Channel Axis Ambiguity",
                        "Is the reference image stored as channels-first (axis 0) or channels-last (last axis)?",
                        ["Channels-first", "Channels-last"],
                        0, False
                    )
                    if not ok:
                        return
                    channel_axis = 0 if choice == "Channels-first" else -1
                elif small_first:
                    channel_axis = 0
                elif small_last:
                    channel_axis = -1
                else:
                    channel_axis = None

                if channel_axis is not None:
                    if channel_axis == 0:
                        channels = ref_img.shape[0]
                        prompt = "Reference image has multiple channels (channels-first). Choose one:"
                    else:
                        channels = ref_img.shape[-1]
                        prompt = "Reference image has multiple channels (channels-last). Choose one:"
                    if channels > 1:
                        channel_str, ok = QtWidgets.QInputDialog.getItem(
                            self, "Select Channel", prompt,
                            [f"Ch. {i+1}" for i in range(channels)], 0, False
                        )
                        if not ok:
                            return
                        chosen_channel = int(channel_str.split()[-1]) - 1
                        if channel_axis == 0:
                            ref_img = ref_img[chosen_channel, :, :]
                        else:
                            ref_img = ref_img[:, :, chosen_channel]
            ref_img = np.squeeze(ref_img)

            # (Optionally, verify that its dimensions match the current movie frame.)
            if self.movie is not None:
                movie_frame = self.movieCanvas.image
                if movie_frame is None:
                    movie_frame = self.get_movie_frame(0)
                if movie_frame.shape[0:2] != ref_img.shape[0:2]:
                    QMessageBox.warning(
                        self,
                        "Dimension Mismatch",
                        "The reference image x/y dimensions do not match the currently displayed movie frame."
                    )
                    return

            self.referenceImage = ref_img
            self.referenceImage_raw = ref_img
            self._ref_dx = 0
            self._ref_dy = 0

            # *** Compute reference contrast settings ***
            p15, p99 = np.percentile(ref_img, (15, 99))
            ref_vmin = int(p15)
            ref_vmax = int(p99 * 1.1)
            delta = ref_vmax - ref_vmin
            self.reference_contrast_settings = {
                'vmin': ref_vmin,
                'vmax': ref_vmax,
                'extended_min': ref_vmin - int(0.7 * delta),
                'extended_max': ref_vmax + int(1.4 * delta)
            }
            # Make the Ref. button visible.
            self.refBtn.setVisible(True)

            reffilt = self.refBtn._bubble_filter
            reffilt._wobj = self.refBtn
            self.refBtn.setChecked(True)
            QTimer.singleShot(1000, lambda: reffilt._showBubble(force=True))

        except Exception as e:
            QMessageBox.critical(self, "Load Error", f"Could not load reference image:\n{str(e)}")


    def on_ref_toggled(self, checked):
        if checked:
            # Turn off sum mode if active.
            if self.sumBtn.isChecked():
                self.sumBtn.setChecked(False)
            # Apply the reference image and its contrast.
            settings = self.reference_contrast_settings
            self.movieCanvas.set_display_range(settings['vmin'], settings['vmax'])

            # Use the raw reference and apply the persistent translation.
            ref_raw = getattr(self, "referenceImage_raw", None)
            if ref_raw is None:
                ref_raw = getattr(self, "referenceImage", None)

            dx = int(getattr(self, "_ref_dx", 0))
            dy = int(getattr(self, "_ref_dy", 0))
            ref_show = self._shift_image_no_wrap(ref_raw, dx, dy, fill_value=0)

            self.movieCanvas.image = ref_show
            if self.movieCanvas._im is not None:
                self.movieCanvas._im.set_data(ref_show)
                
            self.movieCanvas._bg = None
            self._rebuild_movie_blit_background()

        else:
            self.refBtn.setStyleSheet("")
            # Only revert if sum mode is off.
            if not self.sumBtn.isChecked():
                frame = self.get_movie_frame(self.frameSlider.value())
                if frame is not None:
                    # Determine the appropriate contrast settings.
                    if self.movie.ndim == 4:
                        try:
                            current_channel = int(self.movieChannelCombo.currentText())
                        except Exception:
                            current_channel = 1
                        # If the contrast settings haven't been set for the current channel, compute defaults.
                        if current_channel not in self.channel_contrast_settings:
                            p15, p99 = np.percentile(frame, (15, 99))
                            default_vmin = int(p15)
                            default_vmax = int(p99 * 1.1)
                            delta = default_vmax - default_vmin
                            settings = {
                                'vmin': default_vmin,
                                'vmax': default_vmax,
                                'extended_min': default_vmin - int(0.7 * delta),
                                'extended_max': default_vmax + int(1.4 * delta)
                            }
                            self.channel_contrast_settings[current_channel] = settings
                        else:
                            settings = self.channel_contrast_settings[current_channel]
                    else:
                        # For a single-channel (3D) movie, we always use channel 1.
                        if 1 not in self.channel_contrast_settings:
                            p15, p99 = np.percentile(frame, (15, 99))
                            default_vmin = int(p15)
                            default_vmax = int(p99 * 1.1)
                            delta = default_vmax - default_vmin
                            settings = {
                                'vmin': default_vmin,
                                'vmax': default_vmax,
                                'extended_min': default_vmin - int(0.7 * delta),
                                'extended_max': default_vmax + int(1.4 * delta)
                            }
                            self.channel_contrast_settings[1] = settings
                        else:
                            settings = self.channel_contrast_settings[1]
                    # Now apply the contrast to the movie canvas.
                    self.movieCanvas.set_display_range(settings['vmin'], settings['vmax'])
                    self.movieCanvas.image = frame
                    if self.movieCanvas._im is not None:
                        self.movieCanvas._im.set_data(frame)

                    self.movieCanvas._bg = None
                    if getattr(self, "temp_movie_analysis_line", None) is not None:
                        self._rebuild_movie_blit_background()
                    else:
                        self.movieCanvas.draw_idle()

    def _shift_image_no_wrap(self, img, dx, dy, fill_value=0):
        """Shift an image by (dx, dy) without wraparound.

        dx > 0 shifts right, dy > 0 shifts down. Areas moved in are filled.
        Works for 2D arrays and for arrays with trailing channel dimensions.
        """
        if img is None:
            return None
        dx = int(dx)
        dy = int(dy)
        if dx == 0 and dy == 0:
            return img

        out = np.empty_like(img)
        out[...] = fill_value

        h, w = img.shape[0], img.shape[1]

        if dx >= 0:
            src_x0, src_x1 = 0, max(0, w - dx)
            dst_x0, dst_x1 = dx, w
        else:
            src_x0, src_x1 = -dx, w
            dst_x0, dst_x1 = 0, max(0, w + dx)

        if dy >= 0:
            src_y0, src_y1 = 0, max(0, h - dy)
            dst_y0, dst_y1 = dy, h
        else:
            src_y0, src_y1 = -dy, h
            dst_y0, dst_y1 = 0, max(0, h + dy)

        if (src_x1 <= src_x0) or (src_y1 <= src_y0) or (dst_x1 <= dst_x0) or (dst_y1 <= dst_y0):
            return out

        out[dst_y0:dst_y1, dst_x0:dst_x1, ...] = img[src_y0:src_y1, src_x0:src_x1, ...]
        return out

    def _refresh_reference_view(self):
        """Re-render the reference with the current stored translation, if ref mode is active."""
        if not hasattr(self, "refBtn") or not self.refBtn.isChecked():
            return

        ref_raw = getattr(self, "referenceImage_raw", None)
        if ref_raw is None:
            ref_raw = getattr(self, "referenceImage", None)
        if ref_raw is None:
            return

        dx = int(getattr(self, "_ref_dx", 0))
        dy = int(getattr(self, "_ref_dy", 0))
        ref_show = self._shift_image_no_wrap(ref_raw, dx, dy, fill_value=0)

        settings = getattr(self, "reference_contrast_settings", None)
        if settings:
            self.movieCanvas.set_display_range(settings['vmin'], settings['vmax'])

        self.movieCanvas.image = ref_show
        if self.movieCanvas._im is not None:
            self.movieCanvas._im.set_data(ref_show)
        self.movieCanvas.draw_idle()

    def _nudge_reference_translation(self, dx_step, dy_step):
        """Adjust the stored reference translation and redraw if reference is shown."""
        self._ref_dx = int(getattr(self, "_ref_dx", 0)) + int(dx_step)
        self._ref_dy = int(getattr(self, "_ref_dy", 0)) + int(dy_step)
        self._refresh_reference_view()

    def load_kymographs(self):
        fnames, _ = QFileDialog.getOpenFileNames(
            self, "Open Kymograph TIFF(s)", "", "TIFF Files (*.tif *.tiff)"
        )
        if fnames:
            for fname in fnames:
                try:
                    kymo = tifffile.imread(fname)
                    # Check if the kymograph has an invalid shape.
                    if kymo.ndim == 3 and kymo.shape[-1] not in (1, 3, 4):
                        QMessageBox.warning(
                            self, "Invalid Kymograph",
                            f"File '{os.path.basename(fname)}' has an invalid shape {kymo.shape}.\n"
                            "It must be a 2D image or a 3D image with 1, 3, or 4 channels."
                        )
                        continue  # Skip this file.
                        
                    # Generate a unique key for the kymograph.
                    base = os.path.basename(fname)
                    unique_name = base
                    suffix = 1
                    while unique_name in self.kymographs:
                        suffix += 1
                        unique_name = f"{base}-{suffix}"
                    self.kymographs[unique_name] = kymo
                    self.kymoCombo.insertItem(0, unique_name)
                    self.kymoCombo.setEnabled(self.kymoCombo.count() > 0)
                    self.kymoCombo.setCurrentIndex(0)
                except Exception as e:
                    QMessageBox.critical(self, "Error", f"Could not load kymograph {fname}:\n{str(e)}")
        
        self.update_kymo_visibility()

    def load_kymograph_with_overlays(self):
        """
        Load a single kymograph TIFF and its ImageJ multipoint overlay,
        convert into start/end trajectory points, build a DataFrame,
        and hand off to trajectory loader logic.
        """
        fname, _ = QFileDialog.getOpenFileName(
            self,
            "Open Kymograph TIFF", "",
            "TIFF Files (*.tif *.tiff)"
        )
        if not fname:
            return

        try:
            # 1) load image array and validate shape
            kymo = tifffile.imread(fname)
            if kymo.ndim == 3 and kymo.shape[-1] not in (1, 3, 4):
                QMessageBox.warning(
                    self, "Invalid Kymograph",
                    f"File '{os.path.basename(fname)}' has invalid shape {kymo.shape}."
                )
                return

            # store in UI
            base = os.path.basename(fname)
            unique = base
            i = 1
            while unique in self.kymographs:
                i += 1
                unique = f"{base}-{i}"
            self.kymographs[unique] = kymo
            self.kymoCombo.insertItem(0, unique)
            self.kymoCombo.setCurrentIndex(0)

            # 2) extract ROI blob from ImageJ metadata or raw tag
            with tifffile.TiffFile(fname) as tif:
                ij = tif.imagej_metadata or {}
                blob = None
                if 'Overlays' in ij and ij['Overlays']:
                    blob = ij['Overlays'][0]
                elif 'ROI' in ij:
                    blob = ij['ROI']
                else:
                    tag = tif.pages[0].tags.get(50838)
                    blob = tag.value if tag else None

            if blob is None:
                QMessageBox.information(
                    self, "No Overlay",
                    f"No multipoint ROI found in '{base}'."
                )
                return

            # 3) parse blob into list of (x,y)
            points = parse_roi_blob(blob)
            # 4) group into trajectories: two points = one trajectory
            rows = []
            for idx in range(0, len(points), 2):
                sx, sy = points[idx]      # sy is already measured from top
                ex, ey = points[idx+1]
                fs = int(round(sy))
                fe = int(round(ey))
                # map x-axis point back into movie coords
                xs, ys = self.compute_roi_point(self.rois[self.roiCombo.currentText()], sx)
                xe, ye = self.compute_roi_point(self.rois[self.roiCombo.currentText()], ex)
                traj_id = idx//2 + 1
                rows.append({
                    'Trajectory': traj_id,
                    'Frame': fs,
                    'Search Center X': xs,
                    'Search Center Y': ys
                })
                rows.append({
                    'Trajectory': traj_id,
                    'Frame': fe,
                    'Search Center X': xe,
                    'Search Center Y': ye
                })

            df = pd.DataFrame(rows)

            # 5) hand off to a helper that processes a DataFrame
            # You should refactor load_trajectories() into load_trajectories_from_df(df)
            self.trajectoryCanvas.load_trajectories_from_df(df)
            self.kymoCombo.setEnabled(True)

        except Exception as e:
            QMessageBox.critical(self, "Load Error", f"Failed to load overlays: {e}")

        self.update_table_visibility()

    def load_roi(self):
        files, _ = QFileDialog.getOpenFileNames(
            self, "Open ROI File(s)", "", "ROI Files (*.roi *.zip)"
        )
        if not files:
            return

        # 1) Read all ROIs
        rois = {}
        for file in files:
            if file.lower().endswith('.zip'):
                rois.update(read_roi.read_roi_zip(file))
            else:
                rois.update(read_roi.read_roi_file(file))

        # 2) Replace internal ROI store & rebuild the ROI combo
        self.rois = rois
        self.roiCombo.clear()
        for roi_name in sorted(rois):
            self.roiCombo.addItem(roi_name)
        self.roiCombo.setEnabled(bool(rois))
        self.update_roilist_visibility()

        # 3) Ask user if they want to generate kymographs now
        resp = QMessageBox.question(
            self,
            "Generate Kymographs",
            "Generate kymographs for these ROIs across all channels?",
            QMessageBox.Yes | QMessageBox.No
        )

        # 4) Determine how many channels we have
        if self.movie.ndim == 4:
            n_chan = self.movie.shape[self._channel_axis]
        else:
            n_chan = 1

        # 5) For each ROI, loop over ALL channels
        for roi_name, roi in rois.items():
            for ch in range(n_chan):
                kymo_name = f"ch{ch+1}-{roi_name}"
                if resp == QMessageBox.Yes:
                    # Generate kymograph for this channel
                    kymo = self.movieCanvas.generate_kymograph(
                        roi, channel_override=ch+1
                    )
                    self.kymographs[kymo_name] = kymo
                    self.kymo_roi_map[kymo_name] = {
                        "roi":      roi_name,
                        "channel":  ch+1,
                        "orphaned": False
                    }
                else:
                    # Register as orphaned
                    self.kymo_roi_map[kymo_name] = {
                        "roi":      roi_name,
                        "channel":  ch+1,
                        "orphaned": True
                    }

        # 6) Rebuild & show only current channel’s list
        self.update_kymo_list_for_channel()
        self.update_kymo_visibility()

    def _select_next_kymo(self):
        """Advance the kymo combo one step (if possible)."""
        idx = self.kymoCombo.currentIndex()
        if idx >= 0 and idx < self.kymoCombo.count() - 1:
            self.kymoCombo.setCurrentIndex(idx + 1)

    def _select_prev_kymo(self):
        """Go back one step in the kymo combo (if possible)."""
        idx = self.kymoCombo.currentIndex()
        if idx > 0:
            self.kymoCombo.setCurrentIndex(idx - 1)

    def update_kymo_list_for_channel(self):
        ch = int(self.movieChannelCombo.currentText())
        self.kymoCombo.blockSignals(True)
        self.kymoCombo.clear()

        # 1) Populate only this channel’s items
        for name, info in self.kymo_roi_map.items():
            if info["channel"] == ch and not info.get("orphaned", False):
                self.kymoCombo.addItem(name)
        self.kymoCombo.blockSignals(False)

        # Get all names in this channel
        names = [self.kymoCombo.itemText(i) for i in range(self.kymoCombo.count())]

        # 2) If there are no kymographs at all, clear and return
        if not names:
            self.kymoCombo.setCurrentIndex(-1)
            self.kymoCanvas.ax.cla()
            self.kymoCanvas.ax.axis("off")
            self.kymoCanvas.draw_idle()
            # self._last_roi = None
            return

        # 3) Try to find a “sister” matching the last ROI
        sel = None
        last_roi = self._last_roi
        if last_roi is not None:
            for name in names:
                if self.kymo_roi_map[name]["roi"] == last_roi:
                    sel = name
                    break

        # 4) If no sister found, we want a blank canvas
        if sel is None:
            self.kymoCombo.setCurrentIndex(-1)
            self.kymoCanvas.ax.cla()
            self.kymoCanvas.ax.axis("off")
            self.kymoCanvas.draw_idle()
            # don’t change self._last_roi — so that if the user later switches
            # back to a channel where a sister *does* exist, it’ll pop right in.
            return

        # 5) Otherwise select & display the sister
        self.kymoCombo.setCurrentText(sel)
        self.kymo_changed()

    def _save_zoom_for_roi(self, roiName):
        """Stash the current scale & center under this ROI."""
        c = self.kymoCanvas
        self._roi_zoom_states[roiName] = (c.scale, c.zoom_center)

    def _restore_zoom_for_roi(self, roiName):
        """Re-apply stored scale & center (pan) for this ROI, if any."""
        if roiName in self._roi_zoom_states:
            scale, center = self._roi_zoom_states[roiName]
            c = self.kymoCanvas
            c.scale       = scale
            c.zoom_center = center
            c.update_view()

    def kymo_changed(self):

        self.cancel_left_click_sequence()
        
        # — Save last ROI’s view if user did a manual zoom/pan
        if self._last_roi and self.kymoCanvas.manual_zoom:
            self._save_zoom_for_roi(self._last_roi)
            self.kymoCanvas.manual_zoom = False

        # — Grab the new selection
        kymoName = self.kymoCombo.currentText()
        info     = self.kymo_roi_map.get(kymoName)
        if not info:
            # no valid kymo → clear
            self.kymoCanvas.ax.cla()
            self.kymoCanvas.ax.axis("off")
            self.kymoCanvas.draw_idle()
            self._last_roi = None
            return

        roiName = info["roi"]

        # — Sync the ROI & channel controls
        self.roiCombo.setCurrentText(roiName)
        self.movieChannelCombo.blockSignals(True)
        self.movieChannelCombo.setCurrentIndex(info["channel"] - 1)
        self.movieChannelCombo.blockSignals(False)

        # — Display the kymograph
        img = np.flipud(self.kymographs[kymoName])
        self.kymoCanvas.display_image(img)
        self.kymoCanvas.draw_trajectories_on_kymo()
        self.kymoCanvas.draw_idle()

        # — Restore pan+zoom for this ROI (if any)
        self._restore_zoom_for_roi(roiName)

        # — Remember for next save
        self._last_roi = roiName

    def delete_current_kymograph(self):
        current = self.kymoCombo.currentText()
        if not current:
            return

        # 1) Remove mapping and drop any zoom state for its ROI
        mapping = self.kymo_roi_map.pop(current, None)
        if mapping:
            roi_name = mapping["roi"]
            # drop zoom/pan state
            self._roi_zoom_states.pop(roi_name, None)
            # if this was the last ROI we saw, clear it
            if self._last_roi == roi_name:
                self._last_roi = None
            # remove the ROI itself if nobody else references it
            if not any(info["roi"] == roi_name for info in self.kymo_roi_map.values()):
                self.rois.pop(roi_name, None)
                idx = self.roiCombo.findText(roi_name)
                if idx >= 0:
                    self.roiCombo.removeItem(idx)

        # 2) Delete the kymograph
        self.kymographs.pop(current, None)

        # 3) Remove it from the combo
        old_index = self.kymoCombo.currentIndex()
        self.kymoCombo.removeItem(old_index)

        # 4) Show next one or clear
        if self.kymoCombo.count() > 0:
            new_index = old_index - 1 if old_index > 0 else 0
            self.kymoCombo.setCurrentIndex(new_index)
        else:
            self.kymoCanvas.ax.cla()
            self.kymoCanvas.ax.axis("off")
            self.kymoCanvas.draw_idle()

        # 5) Re-run selection & visibility
        self.kymo_changed()
        self.update_kymo_visibility()
        self.update_roilist_visibility()

    def clear_kymographs(self):
        # 1) First, remove any ROIs associated with kymographs
        for mapping in list(self.kymo_roi_map.values()):
            # extract the ROI name whether mapping is a dict or a plain string
            if isinstance(mapping, dict):
                roi_name = mapping.get("roi")
            else:
                roi_name = mapping

            if roi_name and roi_name in self.rois:
                # delete from the internal dict
                del self.rois[roi_name]
                # remove from the ROI combo box
                idx = self.roiCombo.findText(roi_name)
                if idx >= 0:
                    self.roiCombo.removeItem(idx)

        self.kymographs.clear()
        self.kymo_roi_map.clear()
        self._roi_zoom_states.clear()
        self._last_roi = None
        self.kymoCombo.clear()
        self.kymoCanvas.ax.cla()
        self.kymoCanvas.ax.axis("off")
        self.kymoCanvas.draw_idle()
        self.update_kymo_visibility()
        self.update_roilist_visibility()

    def clear_rois(self):
        # Clear the ROI combo box and the dictionary.
        self.rois.clear()
        self.roiCombo.clear()

        # Remove the ROI overlay line (if any) from the movie canvas.
        if hasattr(self.movieCanvas, "roi_line") and self.movieCanvas.roi_line is not None:
            try:
                self.movieCanvas.roi_line.remove()
            except Exception:
                pass
            self.movieCanvas.roi_line = None

        # Also remove any additional ROI lines and text annotations.
        if hasattr(self.movieCanvas, "roi_lines"):
            for line in self.movieCanvas.roi_lines:
                try:
                    line.remove()
                except Exception:
                    pass
            self.movieCanvas.roi_lines = []
        if hasattr(self.movieCanvas, "roi_texts"):
            for txt in self.movieCanvas.roi_texts:
                try:
                    txt.remove()
                except Exception:
                    pass
            self.movieCanvas.roi_texts = []

        self._roi_zoom_states.clear()
        self._last_roi = None
        self.movieCanvas.draw_idle()
        self.update_kymo_visibility()
        self.update_roilist_visibility()
        
    def save_rois(self):
        # Ask user where to save the ZIP file.
        filename, _ = QFileDialog.getSaveFileName(
            self, "Save ROIs as ImageJ ROI Zip", "", "ZIP Files (*.zip)"
        )
        if not filename:
            return

        # Ensure the file has a .zip extension.
        if not filename.lower().endswith('.zip'):
            filename += '.zip'

        try:
            with zipfile.ZipFile(filename, 'w') as zf:
                # Iterate over all ROIs stored in self.rois.
                for roi_name, roi in self.rois.items():
                    # Convert ROI dictionary into ImageJ binary format.
                    roi_bytes = convert_roi_to_binary(roi)
                    file_name = f"{roi_name:03}.roi"
                    zf.writestr(file_name, roi_bytes)
            # Optionally, show a message to the user that the file was saved.
            # QMessageBox.information(self, "Saved", f"ROIs successfully saved to {filename}")
        except Exception as e:
            QMessageBox.critical(self, "Save Error", f"Could not save ROIs:\n{str(e)}")

    def save_kymographs(self):
        # 1) Nothing to save?
        if not self.kymographs:
            QMessageBox.information(self, "No Kymographs", "Nothing to save.")
            return

        # clear any existing selection
        tw = self.trajectoryCanvas.table_widget
        tw.clearSelection()
        tw.setCurrentCell(-1, -1)

        # 2) ask user where/how to save
        all_items = list(self.kymographs.items())
        base_name = os.path.splitext(self.movieNameLabel.text())[0]
        dlg       = SaveKymographDialog(base_name, all_items, parent=self)
        if dlg.exec_() != QDialog.Accepted:
            return
        opts       = dlg.getOptions()
        directory  = opts["directory"]
        sel_names  = opts["selected"]
        ft         = opts["filetype"]
        do_overlay = opts["overlay"]
        use_pref   = opts.get("use_prefix", False)
        mid        = opts.get("middle", "")
        custom     = opts.get("custom", False)
        cname      = opts.get("custom_name", "")

        # 3) progress bar
        total = len(sel_names)
        prog  = QProgressDialog("Saving kymographs…", "Cancel", 0, total, self)
        prog.setWindowModality(Qt.WindowModal)
        prog.setMinimumDuration(0)
        prog.show()

        tw = self.trajectoryCanvas.table_widget
        tw.clearSelection()
        tw.setCurrentCell(-1, -1)

        # remember the current kymo so we can re-select it at the end
        current = self.kymoCombo.currentText()

        # cache figure size for high-DPI save
        fig = self.kymoCanvas.fig
        orig_size = fig.get_size_inches().copy()

        try:
            for i, name in enumerate(sel_names):
                if prog.wasCanceled():
                    break
                prog.setValue(i)

                # build filename
                if custom:
                    fname = cname or name
                else:
                    parts = ([base_name] if use_pref else []) + ([mid] if mid else []) + [name]
                    fname = "-".join(parts)
                out_path = os.path.join(directory, f"{fname}.{ft}")

                if do_overlay:
                    print([name])
                    # 1) load & flip the raw kymo
                    kymo = np.flipud(self.kymographs[name])

                    # 2) ensure display_image does a full reset
                    self.kymoCanvas.manual_zoom = False
                    self.kymoCanvas.display_image(kymo)

                    # 3) switch ROI & channel
                    if name in self.kymographs:
                        self.kymoCombo.setCurrentText(name)
                        self.kymo_changed()

                    # 5) now draw a skinny overlay (axes already off, full‐frame)
                    self.kymoCanvas.draw_trajectories_on_kymo(
                        showsearchline=False,
                        skinny=True
                    )
                    self.kymoCanvas.fig.canvas.draw()

                    # 6) save
                    fig = self.kymoCanvas.fig
                    fig.set_size_inches(orig_size)
                    fig.savefig(out_path, dpi=300,
                                facecolor=fig.get_facecolor(),
                                edgecolor="none",
                                bbox_inches="tight")

                else:
                    # plain export
                    kymo = self.kymographs[name]
                    if ft == "tif":
                        tifffile.imwrite(out_path, kymo, imagej=True)
                    else:
                        p15, p99 = np.percentile(kymo, (15, 99))
                        disp     = np.clip((kymo - p15)/(p99 - p15), 0, 1)
                        disp     = (disp*255).astype(np.uint8)
                        cmap     = "gray_r" if getattr(self, "inverted_cmap", False) else "gray"
                        disp     = np.flipud(disp)
                        plt.imsave(out_path, disp, cmap=cmap, origin="lower")

            prog.setValue(total)

        finally:
            prog.close()
            # just re-select the original kymo; that will reset ROI, channel, contrast, overlays, etc.
            if current in self.kymographs:
                self.kymoCombo.setCurrentText(current)
                self.kymo_changed()

    #UNUSED
    def save_kymograph_with_rois(self):
        """
        Save the selected kymo as a TIFF that ImageJ will open
        with multipoint overlay drawn.
        """

        if not self.kymographs:
            QMessageBox.information(self, "", "Nothing to save.")
            return

        # 1) Ask where to save
        fname, _ = QFileDialog.getSaveFileName(
            self, "Save Kymograph with Overlays", "", "TIFF Files (*.tif *.tiff)"
        )
        if not fname:
            return

        # 2) Grab kymo & movie
        kymo_name = self.kymoCombo.currentText()
        if kymo_name not in self.kymographs:
            QMessageBox.warning(self, "", "Select a kymograph first.")
            return
        kymo = self.kymographs[kymo_name]
        width = kymo.shape[1]
        if self.movie is None:
            QMessageBox.warning(self, "", "Load a movie first.")
            return
        nframes = self.movie.shape[0]

        # 3) Resolve ROI
        roi_key = (
            self.roiCombo.currentText()
            if self.roiCombo.count() > 0
            else kymo_name
        )
        if roi_key not in self.rois:
            QMessageBox.warning(self, "", f"ROI '{roi_key}' not found.")
            return
        roi = self.rois[roi_key]

        # 4) Build overlay points
        pts = []
        for traj in self.trajectoryCanvas.trajectories:
            frames = traj.get("frames", [])
            coords = traj.get('original_coords', [])
            if not frames or not coords:
                continue
            # start
            f0, (x0, y0) = frames[0], coords[0]
            kx0 = self.compute_kymo_x_from_roi(roi, x0, y0, width)
            ky0 = int(round(f0))
            # end
            fn, (xn, yn) = frames[-1], coords[-1]
            kxn = self.compute_kymo_x_from_roi(roi, xn, yn, width)
            kyn = int(round(fn))
            pts.extend([(kx0, ky0), (kxn, kyn)])
        if not pts:
            QMessageBox.information(self, "No Trajectories", "Nothing to save.")
            return

        # 5) Build the ROI blob
        blob = generate_multipoint_roi_bytes(pts)
        print(f"DEBUG: writing ROI blob length {len(blob)} bytes")

        # 6) Build the ImageJ ImageDescription text
        imgdesc = "\n".join([
            "ImageJ=1.53a",
            "images=1",
            "channels=1",
            "slices=1",
            "frames=1",
            "hyperstack=true",
            "overlays=1",
        ]) + "\n"
        desc_bytes = imgdesc.encode('ascii')

        # 7) Write both tags explicitly
        extratags = [
            # tag 270 = ImageDescription
            (270, 's', len(desc_bytes), desc_bytes, True),
            # tag 50838 = ROI
            (50838, 'B', len(blob), blob, True),
        ]
        tifffile.imwrite(
            fname,
            kymo,
            imagej=True,
            metadata={'ROI': blob},  # ← use 'ROI' exactly as ImageJ does
            bigtiff=False
        )

        QMessageBox.information(
            self, "Saved",
            f"Wrote {fname} with {len(pts)//2} trajectories ({len(pts)} points)."
        )

    def show_channel_axis_dialog(self):
        # Check if a movie is loaded and if it is 4-D (with channel options)
        if self.movie is None or self.movie.ndim != 4:
            QMessageBox.information(self, "No extra axes", 
                                    "There are no axes to choose from")
            return

        # Build a list of available axis options (for example, all axes except the time axis)
        # Here we assume axis 0 is time so valid axes are 1, 2, ... movie.ndim-1.
        available_axes = list(range(1, self.movie.ndim))
        dialog = ChannelAxisDialog(available_axes, self)
        if dialog.exec_() == QDialog.Accepted:
            selected_axis = dialog.selected_axis()
            # Set channel axis.
            self._channel_axis = selected_axis
            # Update the channel combo box.
            self.update_movie_channel_combo()

    def set_scale(self):
        # Open the Set Scale dialog prefilled with the current pixel_size and frame_interval values (if any)
        dialog = SetScaleDialog(self.pixel_size, self.frame_interval, parent=self)
        if dialog.exec_() == QDialog.Accepted:
            pixel_size, frame_interval = dialog.get_values()
            if (pixel_size is None) != (frame_interval is None):
                QMessageBox.warning(
                    self,
                    "Incomplete scale",
                    "Both pixel size and frame time are required. Please set both values.",
                )
                return
            self.pixel_size = pixel_size  # in nm
            self.frame_interval = frame_interval  # in ms
            self.update_scale_label()
            self.flash_message("Scale set")
            # Optionally, update any UI elements or print to console:
            #print(f"Set pixel size to {self.pixel_size} nm and frame interval to {self.frame_interval} ms")
            
            # Update velocity information in all trajectories.
            tc = self.trajectoryCanvas  # shortcut to the trajectory canvas
            for row in range(tc.table_widget.rowCount()):
                traj = tc.trajectories[row]
                # Calculate velocities (in pixels per frame)
                velocities = calculate_velocities(traj["spot_centers"])
                valid_velocities = [v for v in velocities if v is not None]
                if valid_velocities:
                    average_velocity = np.mean(valid_velocities)
                else:
                    average_velocity = None
                # Convert average_velocity from pixels/frame to micro meters per second and per minute.
                if self.pixel_size is not None and self.frame_interval is not None and average_velocity is not None:
                    # Here, pixel_size is in nm and frame_interval in ms; the conversion to um/s is:
                    # (average_velocity (px/frame) * pixel_size (nm/px)) / (frame_interval (ms)) 
                    # and then convert nm/ms to um/s by dividing by 1000.
                    velocity_nm_per_ms = (average_velocity * self.pixel_size) / self.frame_interval
                    avg_vel_um_s_txt = f"{velocity_nm_per_ms:.2f}"
                    avg_vel_um_min_txt = f"{velocity_nm_per_ms*60.0:.2f}"
                else:
                    avg_vel_um_s_txt = ""
                    avg_vel_um_min_txt = ""

                dx = traj["end"][1] - traj["start"][1]
                dy = traj["end"][2] - traj["start"][2]
                distance_px = np.hypot(dx, dy)
                time_fr = traj["end"][0] - traj["start"][0]
                distance_um_txt = ""
                time_s_txt = ""
                overall_vel_um_s_txt = ""
                if self.pixel_size is not None and self.frame_interval is not None and time_fr > 0:
                    distance_um = distance_px * self.pixel_size / 1000
                    time_s = time_fr * self.frame_interval / 1000
                    overall_vel_um_s = distance_um/time_s
                    distance_um_txt = f"{distance_um:.2f}"
                    time_s_txt = f"{time_s:.2f}"
                    overall_vel_um_s_txt = f"{overall_vel_um_s:.2f}"

                tc.writeToTable(row, "distance", distance_um_txt)
                tc.writeToTable(row, "time", time_s_txt)                
                tc.writeToTable(row, "netspeed", overall_vel_um_s_txt)
            
            # Update the displayed velocity plot (for the currently selected trajectory)
            selected_rows = tc.table_widget.selectionModel().selectedRows()
            if selected_rows:
                current_row = selected_rows[0].row()
            elif tc.table_widget.rowCount() > 0:
                current_row = 0
                tc.table_widget.selectRow(current_row)
            else:
                current_row = None
            if current_row is not None:
                current_traj = tc.trajectories[current_row]
                self.velocityCanvas.plot_velocity_histogram(current_traj["velocities"])

        self.trajectoryCanvas.hide_empty_columns()

    def correct_drift(self):
        """
        Corrects the drift in the currently loaded movie using spot tracking.

        For multi–channel movies the analysis is performed on the currently selected
        channel but the correction is applied to the full frame so that the original
        movie shape (including channel axis) is preserved. New areas are padded with black.

        The tracking uses the spot center from one frame as the search center for the next.
        If no spot is found, the same displacement as in the previous frame is applied.
        At the end, any gaps are filled in via linear interpolation.
        """
        if self.movie is None:
            QMessageBox.warning(self, "", "Please load a movie first.")
            return

        if not hasattr(self, "drift_reference") or self.drift_reference is None:
            QMessageBox.warning(self, "",
                                "Please click a stationary spot that can be found in all frames first.")
            return

        ref_spot = self.drift_reference  # (x, y)
        n_frames = self.movie.shape[0]
        multi_channel = (self.movie.ndim == 4)

        # Initialize a list to hold the tracked spot centers.
        spot_centers = [None] * n_frames
        spot_centers[self.spot_frame] = ref_spot

        crop_size = int(2 * self.searchWindowSpin.value())

        def get_analysis_frame(full_frame):
            if not multi_channel:
                return full_frame
            current_chan = int(self.movieChannelCombo.currentText()) - 1
            if self._channel_axis == 1:  # channels-first: (channels, H, W)
                return full_frame[current_chan]
            else:  # channels-last: (H, W, channels)
                return full_frame[..., current_chan]

        # --- Create a single progress dialog for both tracking and shifting ---
        total_tracking = (n_frames - self.spot_frame - 1) + self.spot_frame
        total_shifts = n_frames
        total_steps = total_tracking + total_shifts

        progress = QProgressDialog("Tracking spot and applying shift...", "Cancel", 0, total_steps, self)
        progress.setWindowModality(Qt.WindowModal)
        progress.setMinimumDuration(0)
        progress.show()
        current_progress = 0

        # --- Forward Tracking ---
        current_spot = ref_spot
        last_disp = (0, 0)
        for i in range(self.spot_frame + 1, n_frames):
            if progress.wasCanceled():
                return
            full_frame = self.movie[i] if multi_channel else self.get_movie_frame(i)
            if full_frame is None:
                spot_centers[i] = current_spot
                current_progress += 1
                progress.setValue(current_progress)
                continue

            analysis_frame = get_analysis_frame(full_frame)
            analysis_frame = np.atleast_2d(analysis_frame)
            if analysis_frame.ndim != 2:
                spot_centers[i] = current_spot
                current_progress += 1
                progress.setValue(current_progress)
                continue

            try:
                fitted_center, sigma, intensity, peak, bkgr = perform_gaussian_fit(analysis_frame, current_spot, crop_size, pixelsize = self.pixel_size)
            except Exception as e:
                print(f"Forward Gaussian fit error at frame {i}: {e}")
                fitted_center = None

            if fitted_center is not None:
                new_spot = fitted_center
                last_disp = (new_spot[0] - current_spot[0], new_spot[1] - current_spot[1])
            else:
                new_spot = (current_spot[0] + last_disp[0], current_spot[1] + last_disp[1])
            spot_centers[i] = new_spot
            current_spot = new_spot
            current_progress += 1
            progress.setValue(current_progress)

        # --- Backward Tracking ---
        current_spot = ref_spot
        last_disp = (0, 0)
        for i in range(self.spot_frame - 1, -1, -1):
            if progress.wasCanceled():
                return
            full_frame = self.movie[i] if multi_channel else self.get_movie_frame(i)
            if full_frame is None:
                spot_centers[i] = current_spot
                current_progress += 1
                progress.setValue(current_progress)
                continue

            analysis_frame = get_analysis_frame(full_frame)
            analysis_frame = np.atleast_2d(analysis_frame)
            if analysis_frame.ndim != 2:
                spot_centers[i] = current_spot
                current_progress += 1
                progress.setValue(current_progress)
                continue

            try:
                fitted_center, sigma, intensity, peak, bkgr = perform_gaussian_fit(analysis_frame, current_spot, crop_size, pixelsize = self.pixel_size)
            except Exception as e:
                print(f"Backward Gaussian fit error at frame {i}: {e}")
                fitted_center = None

            if fitted_center is not None:
                new_spot = fitted_center
                last_disp = (new_spot[0] - current_spot[0], new_spot[1] - current_spot[1])
            else:
                new_spot = (current_spot[0] + last_disp[0], current_spot[1] + last_disp[1])
            spot_centers[i] = new_spot
            current_spot = new_spot
            current_progress += 1
            progress.setValue(current_progress)
        # End of tracking phase.

        # --- Fill in Gaps with Linear Interpolation ---
        for i in range(n_frames):
            if spot_centers[i] is None:
                prev = i - 1
                while prev >= 0 and spot_centers[prev] is None:
                    prev -= 1
                nxt = i + 1
                while nxt < n_frames and spot_centers[nxt] is None:
                    nxt += 1
                if prev >= 0 and nxt < n_frames:
                    t = (i - prev) / (nxt - prev)
                    x_interp = spot_centers[prev][0] + t * (spot_centers[nxt][0] - spot_centers[prev][0])
                    y_interp = spot_centers[prev][1] + t * (spot_centers[nxt][1] - spot_centers[prev][1])
                    spot_centers[i] = (x_interp, y_interp)
                elif prev >= 0:
                    spot_centers[i] = spot_centers[prev]
                elif nxt < n_frames:
                    spot_centers[i] = spot_centers[nxt]
                else:
                    spot_centers[i] = ref_spot

        displacements = [(sc[0] - ref_spot[0], sc[1] - ref_spot[1]) for sc in spot_centers]

        # --- Apply Correction (Shifting) ---
        corrected_frames = [None] * n_frames

        for i in range(n_frames):
            if progress.wasCanceled():
                return
            full_frame = self.movie[i] if multi_channel else self.get_movie_frame(i)
            if full_frame is None:
                corrected_frames[i] = None
                current_progress += 1
                progress.setValue(current_progress)
                continue

            if full_frame.ndim == 2:
                shift_vector = [-displacements[i][1], -displacements[i][0]]
            elif full_frame.ndim == 3:
                if multi_channel:
                    if self._channel_axis == 1:  # channels-first: (channels, H, W)
                        shift_vector = [0, -displacements[i][1], -displacements[i][0]]
                    else:  # channels-last: (H, W, channels)
                        shift_vector = [-displacements[i][1], -displacements[i][0], 0]
                else:
                    shift_vector = [-displacements[i][1], -displacements[i][0]]
            else:
                QMessageBox.critical(self, "Drift Correction Error",
                                    "Unexpected movie dimensions; cannot apply drift correction.")
                progress.close()
                return

            try:
                # Use order=0 (nearest neighbor) so that pixel values are unchanged.
                corrected_frames[i] = shift(full_frame, shift=shift_vector, order=0, mode='constant', cval=0)
            except Exception as e:
                QMessageBox.critical(self, "Drift Correction Error",
                                    f"Error applying shift on frame {i}:\n{e}")
                corrected_frames[i] = full_frame
            current_progress += 1
            progress.setValue(current_progress)
        progress.close()

        # --- Display the Corrected Movie in a Popup Dialog ---
        dialog = QDialog(self)
        dialog.setWindowTitle("Drift-Corrected Movie")
        dialog_layout = QVBoxLayout(dialog)

        # Create a new MovieCanvas for the dialog:
        corrected_canvas = MovieCanvas(dialog, navigator=self)

        # First, determine the contrast settings for the current channel and mode:
        if self.movie.ndim == 4:
            try:
                current_channel = int(self.movieChannelCombo.currentText())
            except Exception:
                current_channel = 1
            settings = self.channel_contrast_settings.get(current_channel)
        else:
            settings = self.channel_contrast_settings.get(1)
        
        # If settings exist, assign them to corrected_canvas:
        if settings is not None:
            corrected_canvas._default_vmin = settings['vmin']
            corrected_canvas._default_vmax = settings['vmax']
            corrected_canvas._vmin = settings['vmin']
            corrected_canvas._vmax = settings['vmax']

        # Add the corrected_canvas to the dialog.
        dialog_layout.addWidget(corrected_canvas)

        # Create a slider for frame navigation.
        slider = QSlider(Qt.Horizontal)
        slider.setMinimum(0)
        slider.setMaximum(len(corrected_frames) - 1)
        dialog_layout.addWidget(slider)

        # Create a channel dropdown if the movie has multiple channels.
        channel_dropdown = None
        if self.movie.ndim == 4:
            channel_dropdown = QComboBox()
            n_channels = self.movie.shape[self._channel_axis]
            for ch in range(n_channels):
                channel_dropdown.addItem(f"Channel {ch+1}")
            # Set the initial channel to match the main GUI.
            channel_dropdown.setCurrentIndex(int(self.movieChannelCombo.currentText()) - 1)
            dialog_layout.addWidget(channel_dropdown)

        # Define a function to update the displayed frame:
        def update_frame(val):
            frame = corrected_frames[val]
            if frame is None:
                return
            if self.movie.ndim == 4 and channel_dropdown is not None:
                ch_index = channel_dropdown.currentIndex()
                if self._channel_axis == 1:
                    display_frame = frame[ch_index]
                else:
                    display_frame = frame[..., ch_index]
            else:
                display_frame = frame

            corrected_canvas.update_image_data(display_frame)

        slider.valueChanged.connect(update_frame)

        # 2) Now define the channel‐change callback, using the same 1‑based keys
        def on_channel_dropdown(ch0):
            # ch0 is zero-based; settings dict uses 1-based keys.
            chan_key = ch0 + 1

            # ensure we have defaults for this channel
            # get a slice of the first corrected frame
            first_frame = corrected_frames[0]
            if self.movie.ndim == 4:
                if self._channel_axis == 1:
                    sample = first_frame[ch0]
                else:
                    sample = first_frame[..., ch0]
            else:
                sample = first_frame

            if chan_key not in self.channel_contrast_settings and sample is not None:
                p15, p99 = np.percentile(sample, (15, 99))
                vmin = int(p15)
                vmax = int(p99 * 1.1)
                d = vmax - vmin
                self.channel_contrast_settings[chan_key] = {
                    'vmin': vmin,
                    'vmax': vmax,
                    'extended_min': vmin - int(0.7*d),
                    'extended_max': vmax + int(1.4*d)
                }
            settings = self.channel_contrast_settings[chan_key]

            corrected_canvas._default_vmin = settings["vmin"]
            corrected_canvas._default_vmax = settings["vmax"]
            corrected_canvas._vmin = settings["vmin"]
            corrected_canvas._vmax = settings["vmax"]

            # finally, repaint current slider frame with new contrast
            update_frame(slider.value())

        if channel_dropdown is not None:
            channel_dropdown.currentIndexChanged.connect(on_channel_dropdown)

        # 3) Kick it off once at startup
        update_frame(0)

        # Add dialog buttons.
        btn_layout = QHBoxLayout()
        btn_save = QPushButton("Save Movie")
        btn_save_load = QPushButton("Save and Load Movie")
        btn_cancel = QPushButton("Cancel")
        btn_layout.addWidget(btn_save)
        btn_layout.addWidget(btn_save_load)
        btn_layout.addWidget(btn_cancel)
        dialog_layout.addLayout(btn_layout)

        saved_file = {"path": None}

        def save_movie():
            # static file-picker; parent is 'dialog'
            fname, _ = QFileDialog.getSaveFileName(
                dialog,
                "Save Drift-Corrected Movie",
                "",
                "TIFF Files (*.tif *.tiff)"
            )
            if not fname:
                # user clicked Cancel in the file-chooser → do nothing
                return False
            try:
                # actually write it out
                tifffile.imwrite(
                    fname,
                    np.array(corrected_frames),
                    #bigtiff=True, NEEDS TESTING
                    imagej=True,
                    metadata=getattr(self, "movie_metadata", {})
                )
                saved_file["path"] = fname
                return True
            except Exception as e:
                QMessageBox.critical(dialog, "Save Error", f"Error saving movie:\n{e}")
                return False

        def save_and_load_movie():
            # bail out if the user canceled (or if save_movie hit an error)
            if not save_movie():
                return

            # at this point, saved_file["path"] must be set
            try:
                self.save_and_load_routine = True
                self.handle_movie_load(
                    saved_file["path"],
                    pixelsize=self.pixel_size,
                    frameinterval=self.frame_interval
                )
                QMessageBox.information(
                    dialog, "Loaded",
                    "The corrected movie has been loaded into the main window."
                )
                self.zoomInsetFrame.setVisible(False)
            except Exception as e:
                QMessageBox.critical(dialog, "Load Error", f"Error loading movie:\n{e}")
                return

            # only now close the corrected-movie popup
            dialog.accept()

        def cancel():
            dialog.accept()

        def on_save_clicked():
            if save_movie():
                # only close the corrected‐movie popup if we actually saved
                dialog.accept()

        btn_save.clicked.connect(on_save_clicked)
        btn_save_load.clicked.connect(save_and_load_movie)
        btn_cancel.clicked.connect(dialog.reject)

        dialog.exec_()



from ._shared import *

class RecalcDialog(QDialog):
    def __init__(self, current_mode, current_radius, message = "", parent=None):
        super().__init__(parent)
        self.setWindowTitle("Recalculate")
        self.new_mode = current_mode
        self.new_radius = current_radius

        # Main layout
        layout = QVBoxLayout()

        self.setStyleSheet(QApplication.instance().styleSheet())

        # Display message with the number of trajectories needing recalculation.
        message_label = QLabel(message)
        layout.addWidget(message_label)

        # Tracking Mode dropdown
        mode_layout = QHBoxLayout()
        mode_label = QLabel("Tracking mode:")
        mode_label.setAlignment(Qt.AlignRight | Qt.AlignVCenter)
        self.mode_combo = QComboBox()
        self.mode_combo.addItems(["Independent", "Tracked", "Smooth"]) #, "Same center"
        self.mode_combo.setCurrentText(current_mode)
        mode_layout.addWidget(mode_label)
        mode_layout.addWidget(self.mode_combo)
        layout.addLayout(mode_layout)

        # Search radius spin box
        radius_layout = QHBoxLayout()
        radius_label = QLabel("Search Radius:")
        # radius_label.setStyleSheet(
        #     "font-weight: bold"
        # )
        radius_label.setAlignment(Qt.AlignRight | Qt.AlignVCenter)
        self.radius_spin = QSpinBox()
        self.radius_spin.setRange(8, 50)
        self.radius_spin.setValue(current_radius)
        radius_layout.addWidget(radius_label)
        radius_layout.addWidget(self.radius_spin)
        layout.addLayout(radius_layout)

        # OK and Cancel buttons
        button_layout = QHBoxLayout()
        ok_button = QPushButton("Process")
        ok_button.setAutoDefault(True)
        ok_button.setDefault(True)
        cancel_button = QPushButton("Cancel")
        ok_button.clicked.connect(self.accept)
        cancel_button.clicked.connect(self.reject)
        button_layout.addWidget(cancel_button)
        button_layout.addWidget(ok_button)
        layout.addLayout(button_layout)

        self.setLayout(layout)

    def accept(self):
        # Store the chosen values
        self.new_mode = self.mode_combo.currentText()
        self.new_radius = self.radius_spin.value()
        super().accept()

class RecalcWorker(QObject):
    progress = pyqtSignal(int)
    finished = pyqtSignal(list)
    canceled = pyqtSignal()

    def __init__(self, rows, trajectories, navigator):
        super().__init__()
        self._rows         = rows
        self._trajectories = trajectories
        self._navigator    = navigator
        self._is_canceled  = False

    @pyqtSlot()
    def run(self):
        total = sum(len(self._trajectories[r]["frames"]) for r in self._rows)
        count = 0
        results = []
        # If diffusion is enabled, require scale to be set (nm/px and ms)
        if getattr(self._navigator, "show_diffusion", False):
            if getattr(self._navigator, "pixel_size", None) is None or getattr(self._navigator, "frame_interval", None) is None:
                raise ValueError(
                    "Diffusion requires scale: set Pixel size (nm) and Frame interval (ms) before recalculating."
                )
        for row in self._rows:
            if self._is_canceled:
                self.canceled.emit()
                return
            old = self._trajectories[row]
            new_traj = self._navigator._rebuild_one_trajectory(old, self._navigator)

            # Preserve / merge custom_fields
            if "custom_fields" not in new_traj:
                new_traj["custom_fields"] = old.get("custom_fields", {}).copy()
            else:
                merged = old.get("custom_fields", {}).copy()
                merged.update(new_traj.get("custom_fields") or {})
                new_traj["custom_fields"] = merged

            # Optionally recompute diffusion (D, alpha) and store into custom_fields
            if getattr(self._navigator, "show_diffusion", False):
                try:
                    D, alpha = self._navigator.compute_diffusion_for_data(
                        new_traj.get("frames", []),
                        new_traj.get("spot_centers", [])
                    )
                except Exception:
                    D, alpha = (None, None)

                cf = new_traj.setdefault("custom_fields", {})

                # Prefer explicit navigator-provided column names if present
                d_key = (
                    getattr(self._navigator, "_DIFF_D_COL", None)
                    or getattr(self._navigator, "diffusion_D_col", None)
                    or getattr(self._navigator, "diffusion_d_col", None)
                )
                a_key = (
                    getattr(self._navigator, "_DIFF_A_COL", None)
                    or getattr(self._navigator, "diffusion_alpha_col", None)
                    or getattr(self._navigator, "diffusion_a_col", None)
                )

                # Fall back to existing custom_fields keys if they already exist
                if d_key is None or a_key is None:
                    for k in list(cf.keys()):
                        lk = k.lower()
                        if a_key is None and ("alpha" in lk or "α" in k):
                            a_key = k
                        if d_key is None and (lk.strip() == "d" or "diffusion" in lk):
                            if "alpha" not in lk and "α" not in k:
                                d_key = k

                # Final fallbacks
                if d_key is None:
                    d_key = "D"
                if a_key is None:
                    a_key = "alpha"

                # Formatting: keep D compact (4 significant figures) and alpha consistent (4 decimals)
                cf[d_key] = "" if D is None else f"{float(D):.4f}"
                cf[a_key] = "" if alpha is None else f"{float(alpha):.3f}"

            results.append((row, new_traj))

            count += len(old["frames"])
            self.progress.emit(count)

        self.finished.emit(results)

    def cancel(self):
        self._is_canceled = True

class RecalcAllWorker(QObject):
    progress = pyqtSignal(int)        # emits number of frames processed so far
    finished = pyqtSignal(dict)       # emits a dict mapping row → new trajectory
    canceled = pyqtSignal()

    def __init__(self, backup_trajectories: list, navigator):
        super().__init__()
        self._backup = backup_trajectories
        self._navigator = navigator
        # Use exactly one flag for cancellation:
        self._navigator._is_canceled = False

    @pyqtSlot()
    def run(self):
        # print("▶ run() entered; initial _is_canceled =", self._navigator._is_canceled)
        processed = 0
        results = {}
        # If diffusion is enabled, require scale to be set (nm/px and ms)
        if getattr(self._navigator, "show_diffusion", False):
            if getattr(self._navigator, "pixel_size", None) is None or getattr(self._navigator, "frame_interval", None) is None:
                raise ValueError(
                    "Diffusion requires scale: set Pixel size (nm) and Frame interval (ms) before recalculating."
                )

        for row_index, old in enumerate(self._backup):
            # print(f"worker (top of loop) _is_canceled = {self._navigator._is_canceled}")
            # 1) Check the one shared flag at the top of each iteration
            if self._navigator._is_canceled:
                self.canceled.emit()
                return

            # 2) Build “pts” list, checking the same flag inside any nested loops
            if len(old["anchors"]) > 1 and old.get("roi") is not None:
                pts = []
                anchors, roi = old["anchors"], old["roi"]
                for i in range(len(anchors) - 1):
                    if self._navigator._is_canceled:
                        self.canceled.emit()
                        return

                    f1, x1, y1 = anchors[i]
                    f2, x2, y2 = anchors[i+1]
                    seg = range(f1, f2+1) if i == 0 else range(f1+1, f2+1)
                    xs = np.linspace(x1, x2, len(seg), endpoint=True)
                    for j, f in enumerate(seg):
                        if self._navigator._is_canceled:
                            self.canceled.emit()
                            return
                        mx, my = self._navigator.compute_roi_point(roi, xs[j])
                        pts.append((f, mx, my))
            else:
                pts = []
                for f, (x, y) in zip(old["frames"], old["original_coords"]):
                    if self._navigator._is_canceled:
                        self.canceled.emit()
                        return
                    pts.append((f, x, y))

            # 3) Just before calling _compute_analysis, check again
            if self._navigator._is_canceled:
                self.canceled.emit()
                return
            
            # print(f"compute (just before calling _compute_analysis) _is_canceled = {self._navigator._is_canceled}")

            # 4) Run compute_analysis (no GUI), catch exceptions
            try:
                traj_background = self._navigator.compute_trajectory_background(
                    self._navigator.get_movie_frame,
                    pts,
                    crop_size=int(2 * self._navigator.searchWindowSpin.value())
                )
                frames, coords, search_centers, ints, fit, background = (
                    self._navigator._compute_analysis(
                        pts,
                        traj_background,
                        showprogress=False
                    )
                )
            except Exception:
                # skip this trajectory but still bump the progress bar
                processed += len(old["frames"])
                self.progress.emit(processed)
                continue

            # print(f"compute returned; now checking cancellation → {self._navigator._is_canceled}")

            # 5) Immediately after compute_analysis, check cancellation again
            if self._navigator._is_canceled:
                self.canceled.emit()
                return

            # 6) Unpack & rebuild new_traj (same as before)
            spots  = [p[0] for p in fit]
            sigmas = [p[1] for p in fit]
            peaks  = [p[2] for p in fit]
            valid_ints = [v for v, s in zip(ints, spots) if v and v > 0 and s]
            avg_int = float(np.mean(valid_ints)) if valid_ints else None
            med_int = float(np.median(valid_ints)) if valid_ints else None

            vels = []
            for i in range(1, len(spots)):
                p0, p1 = spots[i-1], spots[i]
                if p0 is None or p1 is None:
                    vels.append(None)
                else:
                    vels.append(float(np.hypot(p1[0]-p0[0], p1[1]-p0[1])))
            good_vels = [v for v in vels if v is not None]
            avg_vpf   = float(np.mean(good_vels)) if good_vels else None

            full_centers, full_sigmas, full_peaks, full_ints = [], [], [], []
            for f in old["frames"]:
                if f in frames:
                    idx = frames.index(f)
                    full_centers.append(spots[idx])
                    full_sigmas.append(sigmas[idx])
                    full_peaks.append(peaks[idx])
                    full_ints.append(ints[idx])
                else:
                    full_centers.append(None)
                    full_sigmas.append(None)
                    full_peaks.append(None)
                    full_ints.append(None)

            new_traj = {
                "trajectory_number": old["trajectory_number"],
                "channel":           old["channel"],
                "start":             old["start"],
                "end":               old["end"],
                "anchors":           old["anchors"],
                "roi":               old["roi"],
                "spot_centers":      full_centers,
                "sigmas":            full_sigmas,
                "peaks":             full_peaks,
                "fixed_background":  traj_background,
                "background":        background,
                "frames":            old["frames"],
                "original_coords":   old["original_coords"],
                "search_centers":    search_centers,
                "intensities":       full_ints,
                "average":           avg_int,
                "median":           med_int,
                "velocities":        vels,
                "average_velocity":  avg_vpf
            }

            # 7) Recompute colocalization exactly as before
            if getattr(self._navigator, "check_colocalization", False) and self._navigator.movie.ndim == 4:
                nav = self._navigator
                nav.analysis_frames     = new_traj["frames"]
                nav.analysis_fit_params = list(zip(
                    new_traj["spot_centers"],
                    new_traj["sigmas"],
                    new_traj["peaks"]
                ))
                nav.analysis_channel = new_traj["channel"]
                nav._compute_colocalization(showprogress=False)
                any_list = list(nav.analysis_colocalized)
                by_ch = {
                    ch: list(flags)
                    for ch, flags in nav.analysis_colocalized_by_ch.items()
                }
            else:
                if getattr(self._navigator, "movie", None) is None or self._navigator._channel_axis is None:
                    n_chan = 1
                else:
                    n_chan = self._navigator.movie.shape[self._navigator._channel_axis]

                N = len(new_traj["frames"])
                any_list = [None]*N
                by_ch    = { ch: [None]*N
                             for ch in range(1, n_chan+1)
                             if ch != new_traj["channel"] }

            new_traj["colocalization_any"]   = any_list
            new_traj["colocalization_by_ch"] = by_ch

            # 8) Optionally recompute steps
            if getattr(self._navigator, "show_steps", False):
                idxs, meds = self._navigator.compute_steps_for_data(
                    new_traj["frames"],
                    new_traj["intensities"]
                )
                new_traj["step_indices"] = idxs
                new_traj["step_medians"] = meds
            else:
                new_traj["step_indices"] = None
                new_traj["step_medians"] = None

            # 9) Preserve custom_fields
            new_traj["custom_fields"] = old.get("custom_fields", {}).copy()
            # Optionally recompute diffusion (D, alpha) and store into custom_fields
            if getattr(self._navigator, "show_diffusion", False):
                try:
                    D, alpha = self._navigator.compute_diffusion_for_data(
                        new_traj.get("frames", []),
                        new_traj.get("spot_centers", [])
                    )
                except Exception:
                    D, alpha = (None, None)

                cf = new_traj.setdefault("custom_fields", {})

                # Prefer explicit navigator-provided column names if present
                d_key = (
                    getattr(self._navigator, "_DIFF_D_COL", None)
                    or getattr(self._navigator, "diffusion_D_col", None)
                    or getattr(self._navigator, "diffusion_d_col", None)
                )
                a_key = (
                    getattr(self._navigator, "_DIFF_A_COL", None)
                    or getattr(self._navigator, "diffusion_alpha_col", None)
                    or getattr(self._navigator, "diffusion_a_col", None)
                )

                # Fall back to existing custom_fields keys if they already exist
                if d_key is None or a_key is None:
                    for k in list(cf.keys()):
                        lk = k.lower()
                        if a_key is None and ("alpha" in lk or "α" in k):
                            a_key = k
                        if d_key is None and (lk.strip() == "d" or "diffusion" in lk):
                            if "alpha" not in lk and "α" not in k:
                                d_key = k

                # Final fallbacks
                if d_key is None:
                    d_key = "D"
                if a_key is None:
                    a_key = "alpha"

                # Formatting: keep D compact (4 significant figures) and alpha consistent (4 decimals)
                cf[d_key] = "" if D is None else f"{float(D):.4f}"
                cf[a_key] = "" if alpha is None else f"{float(alpha):.3f}"

            # 10) Store in results and bump progress
            results[row_index] = new_traj
            processed += len(old["frames"])
            self.progress.emit(processed)

        # 11) Finished without cancellation
        #print("▶ run() finished all trajectories without seeing a cancel")
        self.finished.emit(results)

    def cancel(self):
        print("cancel() called, setting flag → True")
        # Called when the user clicks “Cancel” on the QProgressDialog:
        self._navigator._is_canceled = True




from ._shared import *
from .base import ImageCanvas

class MovieCanvas(ImageCanvas):
    def __init__(self, parent=None, navigator=None):
        super().__init__(parent)
        # Initialize panning and display attributes.
        self._is_panning = False
        self._pan_start = None
        self._orig_xlim = None
        self._orig_ylim = None

        self.fig.patch.set_alpha(0)
        self.ax.patch.set_alpha(0)

        self._im = None
        self.image = None
        self._vmin = None
        self._vmax = None
        self._default_vmin = None
        self._default_vmax = None

        self.enableInteraction = True

        self._update_pending = False
        self._inset_update_pending = False
        self._last_inset_params = None

        # New attributes for zooming:
        self.scale = 1.0  # Data units per pixel (uniform in x and y)
        self.padding = 1.25
        self.zoom_center = None  # in data (image) coordinates

        # Connect mouse events.
        self.mpl_connect("scroll_event", self.on_scroll)
        self.mpl_connect("button_press_event", self.on_mouse_press)
        self.mpl_connect("motion_notify_event", self.on_mouse_move)
        self.mpl_connect("button_release_event", self.on_mouse_release)

        self._manual_marker_active = False  # flag for key-controlled marker
        self._manual_marker_pos = None      # will hold [x, y]
        self._manual_marker_artist = None   # store the drawn marker

        self.sum_mode = False
        self.sum_frame_cache = {}

        self._norm_slider_settings = None  # to store normal mode slider settings
        self._sum_slider_settings = None 

        self.tempRoiLine = None
        self._roi_bg = None
        self.roiAddMode = False
        self.roiPoints = [] 

        self.navigator = navigator

        self.last_fitted_center = None
        self.last_fitted_sigma = None
        self.last_intensity_value = None

        self.manual_zoom = False

        self._ctrl_panning = False
        self._last_pan = 0.0

    def mousePressEvent(self, event):
        # Ctrl+Left → pretend it was Middle
        if event.button() == Qt.LeftButton and (event.modifiers() & Qt.ControlModifier):
            self._ctrl_panning = True
            fake = QMouseEvent(
                event.type(),
                event.pos(),
                Qt.MiddleButton,        # button
                Qt.MiddleButton,        # buttons (pressed state)
                event.modifiers()
            )
            super().mousePressEvent(fake)
        else:
            super().mousePressEvent(event)

    def mouseMoveEvent(self, event):
        if self._ctrl_panning:
            fake = QMouseEvent(
                event.type(),
                event.pos(),
                Qt.MiddleButton,
                Qt.MiddleButton if event.buttons() & Qt.LeftButton else Qt.NoButton,
                event.modifiers()
            )
            super().mouseMoveEvent(fake)
        else:
            super().mouseMoveEvent(event)

    def leaveEvent(self, event):
        self._ctrl_panning = False
        super().leaveEvent(event)

    def mouseReleaseEvent(self, event):
        # — finish fake Ctrl+Left pan —
        if self._ctrl_panning and event.button() == Qt.LeftButton:
            # 1) clear both flags
            self._ctrl_panning = False
            self._is_panning    = False

            # 2) send Matplotlib the “middle‐button released” so it can do its cleanup
            fake = QMouseEvent(
                event.type(),
                event.pos(),
                Qt.MiddleButton,
                Qt.NoButton,
                event.modifiers()
            )
            super().mouseReleaseEvent(fake)
            return

        # — finish real middle‐button pan —
        if event.button() == Qt.MiddleButton:
            self._is_panning = False
            super().mouseReleaseEvent(event)
            # do a blocking full draw then snapshot the ROI/marker background
            canvas = self.figure.canvas
            self.draw()                             # synchronous full redraw
            self._roi_bbox = self.ax.bbox           # cache axes bbox
            self._roi_bg   = canvas.copy_from_bbox(self._roi_bbox)
            return

        # — everything else —
        super().mouseReleaseEvent(event)

    def display_image(self, image, title=""):
        """
        Show a 2D image with its native data extent.
        We also initialize our zoom center and scale.
        """
        if image is None:
            return
        self.image = image
        h, w = image.shape
        # Set the image extent to the full image.
        extent = (-0.5, w - 0.5, -0.5, h - 0.5)
        # Initialize zoom_center to the image center if not set yet.
        if self.zoom_center is None:
            self.zoom_center = (w/2, h/2)
        # Set an initial scale such that the entire image is visible.
        # We choose scale so that the data width equals the widget width, unless the widget is not yet sized.
        widget_w = self.width() if self.width() > 0 else w
        widget_h = self.height() if self.height() > 0 else h
        # To show the entire image without zooming, choose the maximum scale needed to cover both dimensions.
        base = max(w / widget_w, h / widget_h)
        self.max_scale = base * self.padding
        self.scale = base
        
        # Draw or update the image.
        if self._im is None:
            self.ax.clear()
            cmap = "gray_r" if getattr(self.navigator, "inverted_cmap", False) else "gray"
            self._im = self.ax.imshow(image, cmap=cmap, vmin=self._vmin, vmax=self._vmax,
                                       origin='lower', extent=extent)
            self.ax.axis("off")
            self.draw()
        else:
            self._im.set_data(image)
            self._im.set_extent(extent)
            self.draw()
            
        # Adjust the view limits based on the new scale and current zoom_center.
        QTimer.singleShot(1, self.update_view)

    def update_image_data(self, image):
        """Update only the image data without changing the current axes limits."""
        if image is None:
            return
        self.image = image
        if self._im is None:
            # if never drawn, fall back
            self.display_image(image)
        else:
            # update pixels…
            self._im.set_data(image)
            # …draw…
            self.draw()
            # …then recapture blit backgrounds so hover/blit won’t restore an old frame
            canvas = self.figure.canvas
            # If navigator is running a blitted temp line, let navigator rebuild a clean bg
            if getattr(self.navigator, "temp_movie_analysis_line", None) is not None:
                self._bg = None
            else:
                self._bg = canvas.copy_from_bbox(self.ax.bbox)
            self._roi_bg = canvas.copy_from_bbox(self.ax.bbox)

    def update_view(self):
        if self.image is None or self.zoom_center is None:
            return

        # 1) compute new limits
        widget_w = self.width()
        widget_h = self.height()
        view_w = widget_w * self.scale
        view_h = widget_h * self.scale
        cx, cy = self.zoom_center
        self.ax.set_xlim(cx - view_w/2, cx + view_w/2)
        self.ax.set_ylim(cy - view_h/2, cy + view_h/2)

        # 2) **blocking** full draw
        self.draw()

        # 3) grab a fresh clean background
        #    (no animated artists in place yet)
        canvas = self.figure.canvas
        # If navigator is running a blitted temp line, let navigator rebuild a clean bg
        if getattr(self.navigator, "temp_movie_analysis_line", None) is not None:
            self._bg = None
        else:
            self._bg = canvas.copy_from_bbox(self.ax.bbox)
        self._roi_bbox = self.ax.bbox
        self._roi_bg   = canvas.copy_from_bbox(self._roi_bbox)

        # reset manual-zoom flag
        self.manual_zoom = False

    def on_scroll(self, event):
        if not self.enableInteraction or self.image is None or event.inaxes != self.ax:
            return
        
        if self.navigator.looping:
            self.manual_zoom = True
            self.navigator.stoploop(prompt=True)
        # 1) grab mouse‐data coords & old scale
        mx, my     = event.xdata, event.ydata
        old_scale  = self.scale
        if mx is None or my is None:
            return

        # 2) compute new scale
        base = 1.2
        if event.button == 'up':
            new_scale = old_scale / base
        else:
            new_scale = old_scale * base

        # clamp to max if set
        if hasattr(self, 'max_scale'):
            new_scale = min(new_scale, self.max_scale)

        # 3) recompute zoom_center so that (mx,my) stays stationary
        cx, cy = self.zoom_center
        ratio  = new_scale / old_scale
        new_cx = mx + (cx - mx) * ratio
        new_cy = my + (cy - my) * ratio

        # 4) store & schedule redraw
        self.scale       = new_scale
        self.zoom_center = (new_cx, new_cy)
        self.update_view()
        # schedule a single zoom/pan update per event loop
    #     if not self._update_pending:
    #         self._update_pending = True
    #         QTimer.singleShot(1, self._perform_throttled_update)

    # def _perform_throttled_update(self):
    #     """
    #     Perform the zoom/pan update in a throttled manner.
    #     """
    #     # full view update then clear the pending flag
    #     self.update_view()
    #     self._update_pending = False
        
    def on_mouse_press(self, event):
        self.manual_zoom = True
        self.navigator._stop_animation = True  
        if not self.enableInteraction:
            return
        if event.inaxes != self.ax:
            return
        if event.button == 2:  # middle-click for panning
            self._is_panning = True
            self._pan_start = (event.x, event.y)
            self._orig_xlim = self.ax.get_xlim()
            self._orig_ylim = self.ax.get_ylim()
            self._last_pan = time.perf_counter()

    def on_mouse_move(self, event):
        if not self._is_panning or event.inaxes != self.ax:
            return
        # throttle pan updates to ~50 Hz
        now = time.perf_counter()
        if now - self._last_pan < 0.02:
            return
        self._last_pan = now

        inv = self.ax.transData.inverted()
        prev_data = inv.transform(self._pan_start)
        current_data = inv.transform((event.x, event.y))
        dx = current_data[0] - prev_data[0]
        dy = current_data[1] - prev_data[1]

        # update the zoom center
        cx, cy = self.zoom_center
        self.zoom_center = (cx - dx, cy - dy)

        # update pan origin for next delta
        self._pan_start = (event.x, event.y)
        self.manual_zoom = True
        
        self.update_view()
        # schedule a single, throttled redraw
        # if not self._update_pending:
        #     self._update_pending = True
        #     QTimer.singleShot(0, self._perform_throttled_update)

    def on_mouse_release(self, event):
        # this is the Matplotlib MouseEvent handler — do NOT call the Qt super()
        if event.button == 2 and event.inaxes == self.ax:
            # finish the pan: full redraw + fresh blit‐background
            self.draw()
            canvas = self.figure.canvas
            self._roi_bbox = self.ax.bbox
            self._roi_bg   = canvas.copy_from_bbox(self._roi_bbox)
            # also update the general bg used for blit‐markers
            self._bg = canvas.copy_from_bbox(self.ax.bbox)
            self.manual_zoom = False

    def resizeEvent(self, event):
        # 1) remember current view center in data coords
        cur_xlim = self.ax.get_xlim()
        cur_ylim = self.ax.get_ylim()
        self.zoom_center = (
            (cur_xlim[0] + cur_xlim[1]) * 0.5,
            (cur_ylim[0] + cur_ylim[1]) * 0.5
        )

        # 2) do the normal resize
        super().resizeEvent(event)
        # Recompute max_scale on resize so zoom-out still fills the canvas
        if hasattr(self, 'image') and self.image is not None:
            h, w = self.image.shape[:2]
            widget_w = self.width() if self.width() > 0 else w
            widget_h = self.height() if self.height() > 0 else h
            base = max(w / widget_w, h / widget_h)
            self.max_scale = base * self.padding

        # 3) redraw using the existing scale & center
        self.update_view()
        
    def update_inset(self, image, center, crop_size, zoom_factor=2,
                    fitted_center=None, fitted_sigma=None,
                    fitted_peak=None, offset=None, intensity_value=None, pointcolor="magenta"):
        # store params
        self._last_inset_params = (image, center, crop_size, zoom_factor,
                                fitted_center, fitted_sigma,
                                fitted_peak, offset, intensity_value, pointcolor)
        
        widget = self.navigator.zoomInsetWidget
        fig = widget.figure

        frame = self.navigator.zoomInsetFrame
        # record default size once
        if not hasattr(self, '_default_inset_size'):
            self._default_inset_size = (frame.width(), frame.height())
        # ensure frame & widget can freely resize
        # frame.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        # self.navigator.zoomInsetWidget.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)

        # ensure 3D axes exists
        if not hasattr(self, 'inset_ax3d'):
            self.inset_ax3d = fig.add_subplot(111, projection='3d')
            self.inset_ax3d.set_axis_off()
            self.inset_ax3d.set_visible(False)

            # hook enter/leave
            cid = fig.canvas
            self._enter_cid = cid.mpl_connect('axes_enter_event', self._on_inset_enter)
            self._leave_cid = cid.mpl_connect('axes_leave_event', self._on_inset_leave)
            self._scroll3d_cid = cid.mpl_connect('scroll_event', self._on_inset_scroll)

        if not getattr(self, '_inset_update_pending', False):
            self._inset_update_pending = True
            # if the inset is currently visible, fire ASAP; otherwise wait 400 ms
            delay = 0 if self.navigator.zoomInsetFrame.isVisible() else 400
            QTimer.singleShot(delay, self._throttled_update_inset)

    def _on_inset_enter(self, event):
        if self.navigator.looping:
            self.navigator.stoploop()
        # if self.navigator.sumBtn.isChecked():
        #     self.navigator.sumBtn.setChecked(False)
        if event.inaxes is self.inset_ax3d or event.inaxes is self.navigator.zoomInsetWidget.ax:
            # hide 2D
            self.navigator.zoomInsetWidget.ax.set_visible(False)
            # show 3D
            self.inset_ax3d.set_visible(True)
            self._draw_threed_inset()

    def _clear_threed_inset(self):
        """Erase the 3D inset and hide its frame."""
        ax3d = self.inset_ax3d
        ax3d.cla()                # clear contents
        ax3d.set_axis_off()       # hide axes
        self.navigator.zoomInsetFrame.setVisible(False)
        self.navigator.zoomInsetWidget.draw()

    def _draw_threed_inset(self):

        """
        The heavy‐lifting routine: crops, zooms, builds the 3D bars +
        Gaussian cap, and then makes everything visible.
        """
        params = self._last_inset_params
        if not params or len(params) != 10:
            return
        (image, center, crop_size, zoom_factor,
        fitted_center, fitted_sigma,
        fitted_peak, offset, intensity_value, pointcolor) = params
        

        # sanity
        if image is None or center is None or np.isnan(center[0]) or np.isnan(center[1]):
            return

        # --- crop & zoom ---
        half = crop_size/2.0
        cx, cy = center
        x1, x2 = cx-half, cx+half
        y1, y2 = cy-half, cy+half
        out_shape = (int(round(y2-y1)), int(round(x2-x1)))
        cropped = subpixel_crop(image, x1, x2, y1, y2, out_shape)
        zoomed = scipy.ndimage.zoom(cropped, zoom_factor, order=0)

        # build/reset axes
        ax3d = self.inset_ax3d
        ax3d.cla()

        # grid & bars
        h,w = zoomed.shape
        x = np.linspace(x1, x2, w);  y = np.linspace(y1, y2, h)
        X,Y = np.meshgrid(x,y)
        xpos, ypos = X.ravel(), Y.ravel()
        # decide on a norm based on the data range:
        dz_raw = zoomed.ravel()
        dz = dz_raw - (float(offset) if offset is not None else 0.0)
        dx,dy = (x2-x1)/w, (y2-y1)/h

        # determine bar bases & heights
        z0     = np.minimum(dz, 0.0)
        height = np.abs(dz)

        dx = (x2 - x1) / w
        dy = (y2 - y1) / h

        # darkest (black) at min(dz), lightest (white) at max(dz)
        norm = mcolors.Normalize(vmin=dz.min(), vmax=dz.max())
        cmap_name = "gray_r" if self.navigator.inverted_cmap else "gray"
        cmap = cm.get_cmap(cmap_name)
        cols = cmap(norm(dz))
        cols[:, 3] = 1  # alpha = 1

        bars = ax3d.bar3d(
            xpos, ypos, z0,
            dx, dy, height,
            color=cols,
            edgecolor='none',
            linewidth=0,
            shade=False
        )
        bars.set_sort_zpos(0.0)

        # Gaussian cap
        if fitted_center is not None and fitted_sigma is not None and fitted_peak is not None:
            x0,y0 = fitted_center;  A = fitted_peak;  σ = fitted_sigma
            G = A*np.exp(-(((X-x0)**2+(Y-y0)**2)/(2*σ**2)))
            z_lift = dz.max()*0 # off for now
            G += z_lift
            mesh_color = pointcolor or "#FF4FA3"
            mesh_rgba = mcolors.to_rgba(mesh_color, 0.65)
            fill_rgba = mcolors.to_rgba(mesh_color, 0.18)

            surf = ax3d.plot_surface(
                X, Y, G,
                color=fill_rgba,
                rstride=1, cstride=1,
                linewidth=0,
                antialiased=True,
                shade=True,
                alpha=fill_rgba[3],
            )
            surf.set_sort_zpos(float(G.max()))

            wf = ax3d.plot_wireframe(
                X, Y, G,
                color=mesh_rgba,
                rstride=2, cstride=2,
                linewidth=0.8,
                alpha=mesh_rgba[3],
            )
            wf.set_zorder(10)
            z_top = G.max()
        else:
            z_top = dz.max()

        # style & zoom
        ax3d.view_init(elev=60, azim=275)
        ax3d.set_axis_off()
        ax3d.set_facecolor((0,0,0,0));  ax3d.set_xlim(x1,x2);  ax3d.set_ylim(y1,y2)
        if not np.isfinite(z_top) or z_top <= 0:
            # fallback to a small positive range, or let Matplotlib autoscale
            try:
                ax3d.autoscale(z=True)
            except Exception:
                pass
        else:
            ax3d.set_zlim(0, z_top)
        for a in (ax3d.xaxis, ax3d.yaxis, ax3d.zaxis):
            a.pane.fill = False;  a.pane.set_edgecolor('none')
        ax3d.grid(False)

        # finally make it visible
        self.navigator.zoomInsetWidget.draw()
        self.navigator.zoomInsetFrame.setVisible(True)

        # update the labels
        if fitted_center is not None:
            self.navigator.zoomInsetLabel.setText(f"{fitted_center[0]:.2f}, {fitted_center[1]:.2f}")
        else:
            self.navigator.zoomInsetLabel.setText("")
        if fitted_center is not None and intensity_value is not None:
            self.navigator.zoomInsetIntensityLabel.setText(f"{intensity_value:.2f}")
        else:
            self.navigator.zoomInsetIntensityLabel.setText("")


    def _on_inset_scroll(self, event):
        if event.inaxes is not self.inset_ax3d:
            return

        frame = self.navigator.zoomInsetFrame
        base   = 1.15
        factor = base if event.button == 'up' else 1/base

        x, y, w, h = frame.geometry().getRect()
        new_w = max(self._default_inset_size[0], min(800, int(w * factor)))
        new_h = max(self._default_inset_size[1], min(800, int(h * factor)))

        # keep top-right fixed
        new_x = x + w - new_w

        # just move + resize the frame
        frame.move(new_x, y)
        frame.resize(new_w, new_h)

        # tell Qt to redo the layout for its children
        frame.layout().invalidate()
        frame.layout().activate()

    def _on_inset_leave(self, event):
        if event.inaxes in (self.inset_ax3d, self.navigator.zoomInsetWidget.ax):
            frame = self.navigator.zoomInsetFrame
            w0, h0 = self._default_inset_size
            geom = frame.geometry()
            new_x = geom.x() + geom.width() - w0

            frame.move(new_x, geom.y())
            frame.resize(w0, h0)
            frame.layout().invalidate()
            frame.layout().activate()

            # existing visibility logic
            self.inset_ax3d.set_visible(False)
            self.navigator.zoomInsetWidget.ax.set_visible(True)
            self._throttled_update_inset()

    def _throttled_update_inset(self):
        """Perform the heavy inset update using the most recent parameters."""
        self._inset_update_pending = False
        params = self._last_inset_params
        if not params or not isinstance(params, tuple) or len(params) !=10:
            return
        
        image, center, crop_size, zoom_factor, fitted_center, fitted_sigma, fitted_peak, offset, intensity_value, pointcolor = self._last_inset_params

        if image is None:
            return
        if center is None or np.isnan(center[0]) or np.isnan(center[1]):
            print("Warning: update_inset received invalid center:", center)
            return
        half = crop_size / 2.0
        x_center, y_center = center[0], center[1]
        x1, x2 = x_center - half, x_center + half
        y1, y2 = y_center - half, y_center + half
        output_shape = (int(round(y2 - y1)), int(round(x2 - x1)))
        cropped = subpixel_crop(image, x1, x2, y1, y2, output_shape) 
        zoomed = scipy.ndimage.zoom(cropped, zoom_factor, order=0)
        self.source_image = image
        self.zoom_extent = (x1, x2, y1, y2)
        
        if hasattr(self.navigator, "zoomInsetWidget"):
            # Update the inset widget’s axes.
            inset_ax = self.navigator.zoomInsetWidget.ax
            inset_ax.clear()
            self.navigator.zoomInsetWidget._im_inset = inset_ax.imshow(
                zoomed,
                cmap=("gray_r" if self.navigator.inverted_cmap else "gray"),
                origin='lower',
                extent=[x1, x2, y1, y2]
            )
            inset_ax.set_xticks([])
            inset_ax.set_yticks([])
            inset_ax.axis('off')

            # Update the overlay text; center it horizontally.
            if fitted_center is not None:
                fc_x, fc_y = fitted_center
                center_text = f"{fc_x:.2f}, {fc_y:.2f}"
            else:
                center_text = ""
            self.navigator.zoomInsetLabel.setText(center_text)
            # Optionally, draw magenta circles if fit parameters are provided.
            if fitted_center is not None and fitted_sigma is not None and intensity_value is not None:
                self.inset_circle = Circle(fitted_center, radius=fitted_sigma * 2, 
                                edgecolor=pointcolor, facecolor='none', linewidth=2, alpha=1)
                inset_ax.add_patch(self.inset_circle)

            if fitted_center is not None and intensity_value is not None:
                self.navigator.zoomInsetIntensityLabel.setText(f"{intensity_value:.2f}")
            else:
                self.navigator.zoomInsetIntensityLabel.setText("")
            self.navigator.zoomInsetWidget.draw()
            # Finally, show the whole zoom inset frame.
            self.navigator.zoomInsetFrame.setVisible(True)
        else:
            self.ax.clear()
            cmap = "gray_r" if getattr(self.navigator, "inverted_cmap", False) else "gray"
            self.ax.imshow(zoomed, cmap=cmap, origin='lower', extent=[x1, x2, y1, y2])
            self.ax.axis('off')

    def update_roi_drawing(self, current_pos):
        pts = list(self.roiPoints) + ([current_pos] if current_pos else [])
        if len(pts) < 2 or not self.roiAddMode:
            return

        xs, ys = zip(*pts)
        canvas = self.figure.canvas     # the QtAgg FigureCanvas

        if self.tempRoiLine is None:
            # 1) draw static image+axes
            canvas.draw()
            # 2) snapshot full axes region (no ROI)
            self._roi_bbox = self.ax.bbox
            self._roi_bg   = canvas.copy_from_bbox(self._roi_bbox)
            # 3) create the line artist (but don’t redraw full figure)
            self.tempRoiLine, = self.ax.plot(xs, ys, '--', linewidth=1.5, color='#81C784')
        else:
            # restore the clean background
            canvas.restore_region(self._roi_bg)
            # update the line
            self.tempRoiLine.set_data(xs, ys)
            # draw just that artist
            self.ax.draw_artist(self.tempRoiLine)
            # blit only the axes region
            canvas.blit(self._roi_bbox)

    def finalize_roi(self):
        # Make sure we have at least two pointsf
        if not self.roiPoints or len(self.roiPoints) < 2:
            print("Not enough points to finalize ROI.")
            return
        
        #print("ROI points:", self.roiPoints)

        # Build the ROI dictionary using all collected points.
        # Use 'x' and 'y' keys expected by the conversion function.
        # and also store the full list as 'points' for any later processing.
        roi = {
            "type": "line",  # or "segmented_line"
            "x": [pt[0] for pt in self.roiPoints],
            "y": [pt[1] for pt in self.roiPoints],
            "points": self.roiPoints.copy()
        }

        # Generate the kymograph from the full ROI.
        kymo = self.generate_kymograph(roi)

        # Combine keys from both dictionaries.
        all_names = set(self.navigator.rois.keys()) | set(self.navigator.kymographs.keys())
        numeric_names = []
        for name in all_names:
            try:
                # Only append if conversion to int is possible.
                numeric_names.append(int(name))
            except (ValueError, TypeError):
                # If a key isn’t numeric, skip it.
                pass

        if numeric_names:
            max_num = max(numeric_names)
            next_num = max_num + 1
        else:
            next_num = 1
        name = f"{next_num:03d}"        

        # Store the ROI and kymograph with the same name.
        self.navigator.rois[name] = roi
        self.navigator.roiCombo.addItem(name)
        self.navigator.roiCombo.setEnabled(True)
        self.navigator.roiCombo.setCurrentText(name)
        self.navigator.update_roilist_visibility()

        if self.navigator.movie.ndim == 4:
            n_chan = self.navigator.movie.shape[self.navigator._channel_axis]
        else:
            n_chan = 1

        for ch in range(n_chan):
            kymo = self.generate_kymograph(roi, channel_override=ch)
            kymo_name = f"ch{ch+1}-{name}"
            self.navigator.kymographs[kymo_name] = kymo
            self.navigator.kymo_roi_map[kymo_name] = {
                "roi":      name,
                "channel":  ch+1,
                "orphaned": False
            }

            # self.navigator.last_kymo_by_channel[ch+1] = kymo_name

        self.navigator._last_roi = name
        self.navigator.update_kymo_list_for_channel()
        self.navigator.kymo_changed()
        self.navigator.kymoCombo.setEnabled(True)


        # Clear the temporary ROI markers and the stored points.
        self.roiPoints = []
        if self.tempRoiLine is not None:
            try:
                self.tempRoiLine.remove()
            except Exception:
                pass
            self.tempRoiLine = None

        self.navigator.update_kymo_visibility()
        self.navigator.update_kymo_list_for_channel()

        self.draw()

    def generate_kymograph(self, roi, channel_override=None):
        # --- Compute the ROI sample positions along the drawn line ---
        xs = np.array(roi["x"], dtype=float)
        ys = np.array(roi["y"], dtype=float)
        
        # Compute cumulative distances along the ROI.
        distances = np.sqrt(np.diff(xs)**2 + np.diff(ys)**2)
        cum_dist = np.concatenate(([0], np.cumsum(distances)))
        total_length = cum_dist[-1]
        
        # Determine the number of sample positions (at least 2)
        num_samples = max(int(total_length), 2)
        sample_positions = np.linspace(0, total_length, num_samples)
        
        # Interpolate the ROI coordinates at these positions.
        sample_x = np.interp(sample_positions, cum_dist, xs)
        sample_y = np.interp(sample_positions, cum_dist, ys)
        
        # --- Compute the tangent and then the normal direction at each sample point ---
        # Use finite differences (np.gradient) to approximate the tangent.
        tangent_dx = np.gradient(sample_x)
        tangent_dy = np.gradient(sample_y)
        # A unit normal can be taken as (-dy, dx)
        normal_x = -tangent_dy
        normal_y = tangent_dx
        norm = np.sqrt(normal_x**2 + normal_y**2)
        norm[norm == 0] = 1  # avoid division by zero if any
        normal_x /= norm
        normal_y /= norm

        # --- Obtain integration parameters ---
        # n pixels in either direction perpendicular to the line
        line_width = getattr(self.navigator, "line_width", 2)
        # Method: "max" (default) or "average"
        line_method = getattr(self.navigator, "line_integration_method", "max").lower()

        # Create an array of offsets along the normal direction.
        # This gives a 1D array from -line_width to +line_width.
        offsets = np.arange(-line_width, line_width + 1, dtype=float)  # shape (n_offsets,)
        n_offsets = offsets.size

        # For each sample point, compute its neighborhood along the normal.
        # These arrays have shape (num_samples, n_offsets).
        sample_x_full = sample_x[:, None] + normal_x[:, None] * offsets[None, :]
        sample_y_full = sample_y[:, None] + normal_y[:, None] * offsets[None, :]

        # Flatten the coordinate arrays so they can be passed to map_coordinates.
        # Note: the first row are y coordinates (rows) and the second row x coordinates.
        coords = np.vstack((sample_y_full.ravel(), sample_x_full.ravel()))

        # --- Retrieve the movie ---
        if hasattr(self, "navigator") and self.navigator.movie is not None:
            movie = self.navigator.movie
        else:
            return None

        # --- Process each frame using vectorized interpolation per frame ---
        kymo_rows = []
        n_frames = movie.shape[0]
        for i in range(n_frames):
            # For multi–channel movies, extract the 2D frame for the chosen channel.
            if movie.ndim == 4:
                frame = movie[i]
                if hasattr(self.navigator, "movieChannelCombo") and self.navigator.movieChannelCombo.isEnabled():
                    if channel_override is not None:
                        channel_index = channel_override
                    else:
                        channel_index = int(self.navigator.movieChannelCombo.currentText()) - 1
                else:
                    channel_index = 0
                if self.navigator._channel_axis == 1:
                    frame_2d = frame[channel_index]
                else:
                    frame_2d = frame[..., channel_index]
            else:
                # For 3D movies (single channel), take the frame directly.
                frame_2d = movie[i]

            # Use map_coordinates to extract the pixel values at all normal offsets,
            # in one vectorized call.
            patch_values = map_coordinates(frame_2d, coords, order=1, mode='reflect')
            # Reshape so that each row corresponds to one sample point (along the ROI)
            # and each column corresponds to one offset along the normal.
            patch_values = patch_values.reshape(num_samples, n_offsets)

            # Integrate along the normal direction based on the selected method.
            if line_method == "average":
                profile = np.mean(patch_values, axis=1)
            else:
                profile = np.max(patch_values, axis=1)
            kymo_rows.append(profile)

        kymo = np.vstack(kymo_rows)

        if self.navigator.applylogfilter:

            # kymo is the generated 2D NumPy array
            sigma = getattr(self.navigator, 'log_sigma', 1.5)

            # convert to float to avoid clipping
            kymo_f = kymo.astype(np.float32)

            # apply LoG (invert for positive edges)
            log_kymo = -gaussian_laplace(kymo_f, sigma=sigma)

            # normalize back to 1–255
            minv, maxv = log_kymo.min(), log_kymo.max()
            if maxv > minv:
                log_kymo = (log_kymo - minv) / (maxv - minv) * 254 + 1
            else:
                log_kymo = np.ones_like(log_kymo) * 128
            log_kymo = log_kymo.astype(np.uint8)

            # from bm3d import bm3d
            # from skimage.restoration import estimate_sigma
            # sigma_est = estimate_sigma(
            #     log_kymo,
            #     channel_axis=None,     # grayscale
            #     average_sigmas=True    # get one scalar back
            # )
            # denoised = bm3d(log_kymo, sigma_psd=sigma_est)

            # # return log_kymo
            # return overlay_trace_centers(denoised)
        
            return(log_kymo)
        
        else:

            return kymo

    def clear_temporary_roi_markers(self):
        # Clear any temporary ROI dotted line.
        if hasattr(self, 'tempRoiLine') and self.tempRoiLine is not None:
            try:
                self.tempRoiLine.remove()
            except Exception:
                pass
            self.tempRoiLine = None

        self.draw()

    def display_sum_frame(self):
        if self.navigator is None or self.navigator.movie is None:
            return

        movie = self.navigator.movie

        # figure out which channel key to use
        if movie.ndim == 4:
            try:
                ch = int(self.navigator.movieChannelCombo.currentText()) - 1
            except Exception:
                ch = 0

            if ch in self.sum_frame_cache:
                sum_frame = self.sum_frame_cache[ch]
            else:
                channel_axis = self.navigator._channel_axis
                idx = [slice(None)] * movie.ndim
                idx[channel_axis] = ch
                channel_movie = movie[tuple(idx)]
                sum_frame = np.max(channel_movie, axis=0)
                self.sum_frame_cache[ch] = sum_frame

        elif movie.ndim == 3:
            if movie.shape[0] <= 4:
                sum_frame = movie[0]
            else:
                sum_frame = np.max(movie, axis=0)
        else:
            sum_frame = movie

        # now render exactly like before
        self.image = sum_frame
        if self._im is None:
            self.ax.clear()
            self._im = self.ax.imshow(sum_frame, cmap="gray", origin='lower')
            self.ax.axis("off")
            self.draw()
        else:
            self._im.set_data(sum_frame)
            self.draw()

        # ── recapture blit backgrounds so future blits use the sum‐mode image ──
        canvas = self.figure.canvas
        # If navigator is running a blitted temp line, let navigator rebuild a clean bg
        if getattr(self.navigator, "temp_movie_analysis_line", None) is not None:
            self._bg = None
        else:
            self._bg = canvas.copy_from_bbox(self.ax.bbox)
        self._roi_bg = canvas.copy_from_bbox(self.ax.bbox)

    def clear_sum_cache(self, channel=None):
        """
        If channel is None, flush everything (e.g. on new movie).
        Otherwise just remove that one channel’s cache.
        """
        if channel is None:
            self.sum_frame_cache.clear()
        else:
            self.sum_frame_cache.pop(channel, None)

    def set_display_range(self, vmin, vmax):
        """
        Set the current display contrast range without modifying underlying data.
        Ensures that vmin is always less than vmax to avoid errors in normalization.
        """
        #print("displaying:", vmin, vmax)
        if vmin >= vmax:
            # If vmin is not less than vmax, adjust vmax to guarantee a valid range.
            # Here we choose an arbitrary minimal gap of 1.
            #print(f"Warning: vmin ({vmin}) >= vmax ({vmax}). Adjusting vmax to {vmin + 1}.")
            vmax = vmin + 1

        self._vmin = vmin
        self._vmax = vmax
        if self._im is not None:
            self._im.set_clim(self._vmin, self._vmax)
            self.draw_idle()

    def overlay_rectangle(self, cx, cy, size, color='#7da1ff'):
        # remove old
        if getattr(self, "rect_overlay", None) is not None:
            try:    self.rect_overlay.remove()
            except: pass

        half = size / 2.0
        x0, x1 = cx - half, cx + half
        y0, y1 = cy - half, cy + half

        # a closed polyline (4 corners + back to first)
        verts_x = [x0, x1, x1, x0, x0]
        verts_y = [y0, y0, y1, y1, y0]

        line, = self.ax.plot(
            verts_x,
            verts_y,
            color=color,
            linewidth=2,
            zorder=6,
            linestyle='-'
        )
        self.rect_overlay = line

    def is_zoomed_in(self):
        if self.image is None or not hasattr(self, "full_extent"):
            return False
        full_left, full_right, full_bottom, full_top = self.full_extent
        cur_xlim = self.ax.get_xlim()
        cur_ylim = self.ax.get_ylim()
        tol = 1e-6
        # Check if current limits are nearly the same as the full extent.
        return not (abs(cur_xlim[0] - full_left) < tol and abs(cur_xlim[1] - full_right) < tol and
                    abs(cur_ylim[0] - full_bottom) < tol and abs(cur_ylim[1] - full_top) < tol)

    def clear_canvas(self):
        """Clear the canvas by removing all overlays and resetting internal state."""
        # Clear the axes.
        self.ax.cla()
        # Remove stored image, marker, and any overlay objects.
        self._im = None
        self._marker = None
        # Clear any gaussian circle if present.
        self.remove_gaussian_circle()
        # Reset panning and manual zoom state.
        self._is_panning = False
        self._pan_start = None
        self._orig_xlim = None
        self._orig_ylim = None
        self.manual_zoom = False
        self.image = None

    def draw_manual_marker(self):
        """Draw a translucent circle at the current manual position."""
        # Remove any existing manual marker
        if getattr(self, "_manual_marker_artist", None) is not None:
            try:
                self._manual_marker_artist.remove()
            except Exception:
                pass

        # Draw a semi‑transparent circle
        x, y = self._manual_marker_pos
        radius = 3  # adjust as desired
        circ = Circle(
            (x, y),
            radius=radius,
            edgecolor='#7da1ff',
            facecolor='#7da1ff',
            alpha=0.6,
            linewidth=1.5
        )
        self._manual_marker_artist = circ
        self.ax.add_patch(circ)

    def clear_manual_marker(self):
        """Remove the manual marker circle from the canvas."""
        if getattr(self, "_manual_marker_artist", None) is not None:
            try:
                self._manual_marker_artist.remove()
            except Exception:
                pass
            self._manual_marker_artist = None

    def add_gaussian_circle(self, fitted_center, fitted_sigma, color="magenta"):
        if fitted_center is not None and fitted_sigma is not None:
            self.gaussian_circle = Circle(
                    fitted_center,
                    radius=2 * fitted_sigma,
                    edgecolor=color,
                    facecolor='none',
                    linewidth=2
                )
            self.ax.add_patch(self.gaussian_circle)

    def remove_gaussian_circle(self):
        removed = False
        if hasattr(self, "gaussian_circle") and self.gaussian_circle is not None:
            try:
                self.gaussian_circle.remove()
                removed = True
            except Exception as e:
                print("Error removing gaussian circle:", e)
            self.gaussian_circle = None
        return removed

    def draw_trajectories_on_movie(self):

        # 1) Clear any existing movie‐canvas markers
        self.clear_movie_trajectory_markers()

        # 2) If the overlay toggle is off, do nothing
        if not self.navigator.traj_overlay_button.isChecked():
            return

        # 3) Which table row is currently selected?
        selected_idx = self.navigator.trajectoryCanvas.table_widget.currentRow()

        # 4) Which movie channel is active? (so we skip mismatched trajectories)
        try:
            current_ch = int(self.navigator.movieChannelCombo.currentText())
        except (ValueError, AttributeError):
            current_ch = None

        # 5) Loop over all trajectories and draw them
        for idx, traj in enumerate(self.navigator.trajectoryCanvas.trajectories):
            # 5a) Skip if trajectory has a channel that doesn't match
            traj_ch = traj.get("channel", None)
            if traj_ch is not None and traj_ch != current_ch:
                continue

            # 5b) Are we highlighting this one?
            is_hl = (idx == selected_idx)

            # Build a label string like "3A"/"3B"
            traj_label = traj.get("file_index") or str(traj["trajectory_number"])

            # 5c) Draw the dashed "search‐center" line if original_coords exist
            original_coords = traj.get("original_coords", [])
            if original_coords:
                xs = [pt[0] for pt in original_coords]
                ys = [pt[1] for pt in original_coords]

                lw_search = 2.0 if is_hl else 1.5
                alpha_search = 0.9 if is_hl else 0.6
                z_search = 5 if is_hl else 1

                dotted_line, = self.ax.plot(
                    xs, ys,
                    color='#7da1ff',
                    linestyle='--',
                    linewidth=lw_search,
                    alpha=alpha_search,
                    zorder=z_search,
                    solid_capstyle='round',
                    dash_capstyle='round'
                )
                self.movie_trajectory_markers.append(dotted_line)

                # 5c-i) Annotate "A" at first point and "B" at last point, picker=True
                dispA = self.ax.transData.transform((xs[0], ys[0]))
                dispB = self.ax.transData.transform((xs[-1], ys[-1]))
                v = dispB - dispA
                norm = (v[0]**2 + v[1]**2) ** 0.5
                u = (v / norm) if norm else np.array([1.0, 0.0])
                offset_px = 15

                for (cx, cy, suffix), sign in [((xs[0], ys[0], 'A'), -1),
                                               ((xs[-1], ys[-1], 'B'), +1)]:
                    dx, dy = u * (offset_px * sign)
                    lbl = self.ax.annotate(
                        f"{traj_label}{suffix}",
                        xy=(cx, cy),
                        xytext=(dx, dy),
                        textcoords="offset points",
                        color=('white' if is_hl else 'black'),
                        fontsize=8,
                        fontweight="bold",
                        ha="center",
                        va="center",
                        bbox=dict(
                            boxstyle='circle,pad=0.3',
                            facecolor=('#7da1ff' if is_hl else '#cbd9ff'),
                            alpha=(0.9 if is_hl else 0.6),
                            linewidth=(1.5 if is_hl else 1.0)
                        ),
                        picker=True,      # Make the label clickable
                        zorder=(7 if is_hl else 2)
                    )
                    # Attach custom attribute so pick_event tells us which row was clicked:
                    lbl.traj_idx = idx
                    self.movie_trajectory_markers.append(lbl)

            # 5d) Draw the solid connecting line through spot_centers for every trajectory
            spot_centers = traj.get('spot_centers', [])
            xs_pts = [pt[0] if pt is not None else np.nan for pt in spot_centers]
            ys_pts = [pt[1] if pt is not None else np.nan for pt in spot_centers]

            scatter_kwargs, line_color = self.navigator._get_traj_colors(traj)

            # Remove any pre‐existing zorder in scatter_kwargs so we can supply our own
            scatter_kwargs = scatter_kwargs.copy()
            scatter_kwargs.pop('zorder', None)

            # Style for the connecting line
            lw_line = (2.0 if is_hl else 1.5)
            alpha_line = (0.9 if is_hl else 0.7)
            z_line = (6 if is_hl else 3)

            line, = self.ax.plot(
                xs_pts, ys_pts,
                linestyle='-',
                color=(line_color),
                linewidth=lw_line,
                alpha=alpha_line,
                zorder=z_line
            )
            self.movie_trajectory_markers.append(line)

            # 5e) Draw scatter points ONLY for the highlighted trajectory
            if is_hl:
                # Bump size and add black edge
                scatter_kwargs.update(s=15, edgecolors='black', linewidths=0.5)
                z_scatter = 6
                scatter = self.ax.scatter(
                    xs_pts, ys_pts,
                    picker=True,
                    zorder=z_scatter,
                    **scatter_kwargs
                )
                scatter.traj_idx = idx
                self.movie_trajectory_markers.append(scatter)

        # 6) Finally, request a redraw
        self.ax.figure.canvas.draw_idle()

    def clear_movie_trajectory_markers(self):
        if hasattr(self, "movie_trajectory_markers"):
            for marker in self.movie_trajectory_markers:
                try:
                    marker.remove()
                except Exception:
                    pass
            self.movie_trajectory_markers = []
        else:
            self.movie_trajectory_markers = []
    
    def remove_inset_circle(self):
        if hasattr(self, "inset_circle"):
            try:
                self.inset_circle.remove()
            except Exception as e:
                print("Error removing inset circle during invalidation:", e)
            self.inset_circle = None



from ._shared import *

class ImageCanvas(FigureCanvas):
    def __init__(self, parent=None):
        self.fig = Figure()
        self.ax = self.fig.add_subplot(111)
        self.fig.subplots_adjust(left=0, right=1, top=1, bottom=0)
        self.ax.axis("off")
        super().__init__(self.fig)
        self.setParent(parent)
        # self.setAttribute(Qt.WA_OpaquePaintEvent)
        # self.setAttribute(Qt.WA_NoSystemBackground)
        self.image = None

    def paintEvent(self, event):
        painter = QPainter(self)
        painter.setClipRect(self.rect())  # This restricts drawing to the visible area.
        super().paintEvent(event)

    def display_image(self, image, title=""):
        self.ax.clear()
        self.image = image
        self.ax.imshow(image, cmap="gray")
        self.ax.set_title(title)
        self.ax.axis('off')
        self.draw()

    def set_cmap(self, cmap):
        # if somebody’s already painted an image into `self._im` or `self.image`:
        im = getattr(self, "_im", None) or getattr(self, "image", None)
        # If AxesImage is stored in `self._im`, use that.
        if hasattr(self, "_im") and self._im is not None:
            self._im.set_cmap(cmap)
            self.draw()



from PyQt5.QtWidgets import (
    QWidget, QVBoxLayout, QHBoxLayout, QPushButton,
    QLabel, QComboBox, QSpinBox, QSplitter,
    QAbstractButton, QFrame, QDialogButtonBox,
    QLineEdit, QFormLayout, QSplitterHandle, QDialog,
    QApplication, QCheckBox, QListWidget, QFileDialog,
    QListWidgetItem, QAbstractItemView, QStackedWidget
)

from PyQt5 import QtCore, QtWidgets, QtGui
from PyQt5.QtCore import (
    Qt, QObject, pyqtSignal, QSize, QRectF, pyqtProperty, QPropertyAnimation,
    QEasingCurve, QObject, QPropertyAnimation, pyqtSlot, QEvent, QRect, QPointF,
    QTimer, QPoint, QElapsedTimer
    )
from PyQt5.QtGui import (
    QPainter, QColor, QFont, QDoubleValidator, QPen, QBrush, QFontMetrics, QPainterPath,
    QCursor, QPolygonF
    )

import numpy as np
import os
from scipy.ndimage import map_coordinates

def subpixel_crop(image, x1, x2, y1, y2, output_shape):
    # Create a grid in the output coordinate system.
    out_y, out_x = np.indices(output_shape, dtype=np.float64)
    # Map the output coordinates back to the input coordinate system.
    # For x: from 0 to output_shape[1]-1 should map to [x1, x2]
    # For y: from 0 to output_shape[0]-1 should map to [y1, y2]
    in_x = x1 + (x2 - x1) * (out_x / (output_shape[1] - 1))
    in_y = y1 + (y2 - y1) * (out_y / (output_shape[0] - 1))
    # Sample the input image at these coordinates.
    return map_coordinates(image, [in_y, in_x], order=1, mode='nearest')



import numpy as np
import os
import pandas as pd
from PyQt5.QtCore import Qt, QTimer, QThread, QEvent
from PyQt5.QtWidgets import (QVBoxLayout, QApplication, QDialog,
                             QWidget, QFileDialog, QMessageBox, QTableWidget,
                             QTableWidgetItem, QMessageBox, QProgressDialog,
                             QHeaderView, QMenu, QInputDialog, QLineEdit, QLabel)
from PyQt5.QtGui import QPainter,QMouseEvent
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure
from matplotlib.patches import Circle, Rectangle
from matplotlib.ticker import MaxNLocator
from matplotlib.lines import Line2D
from matplotlib.text import Text
import matplotlib.cm as cm
import matplotlib.colors as mcolors
from matplotlib.gridspec import GridSpec
from matplotlib.transforms import Bbox
import scipy
from scipy.ndimage import map_coordinates, gaussian_laplace
import time
import copy
import math
import json
import warnings
import re
from typing import Optional, List
from ..tools.roi_tools import is_point_near_roi, compute_roi_point
from ..canvas_tools import RecalcDialog, RecalcWorker, RecalcAllWorker, subpixel_crop
from ..tools.gaussian_tools import filterX, find_minima, find_maxima
# from .kymotrace import prune_skeleton, overlay_trace_centers, extract_main_path

warnings.filterwarnings(
    "ignore",
    message=".*layout engine that is incompatible with subplots_adjust and/or tight_layout.*",
    category=UserWarning,
)
warnings.filterwarnings(
    "ignore",
    message=".*constrained_layout not applied because axes sizes collapsed to zero.*",
    category=UserWarning,
)
warnings.filterwarnings(
    "ignore",
    message=".*Attempting to set identical low and high xlims makes transformation singular.*"
)
warnings.filterwarnings(
    "ignore",
    message=".*Tight layout not applied\\. The left and right margins cannot be made large enough to accommodate all Axes decorations\\.*",
    category=UserWarning,
)
warnings.filterwarnings(
    "ignore",
    message=".*No artists with labels found to put in legend.*",
    category=UserWarning,
)



from ._shared import *

class AxesRectAnimator(QObject):
    def __init__(self, ax):
        super().__init__()
        self._ax = ax

    def getAxesRect(self):
        # Get the current x and y limits of the axes.
        xmin, xmax = self._ax.get_xlim()
        ymin, ymax = self._ax.get_ylim()
        return QRectF(xmin, ymin, xmax - xmin, ymax - ymin)

    def setAxesRect(self, rect):
        # Update the axes limits using the QRectF values.
        self._ax.set_xlim(rect.x(), rect.x() + rect.width())
        self._ax.set_ylim(rect.y(), rect.y() + rect.height())
        # Redraw the canvas.
        self._ax.figure.canvas.draw_idle()

    axesRect = pyqtProperty(QRectF, fget=getAxesRect, fset=setAxesRect)




from ._shared import *

class HistogramCanvas(FigureCanvas):

    def __init__(self, navigator, parent=None):
        self.navigator = navigator
        self.fig = Figure(figsize=(5.5, 3), constrained_layout=True)
        self.ax = self.fig.add_subplot(111)
        self.fig.patch.set_facecolor("white")
        self.ax.set_facecolor("white")
        super().__init__(self.fig)
        self.setParent(parent)
        self.current_background = None
        # Initially, hide the axes to appear blank.
        self.ax.axis("off")
        self.draw()
        self.fig.patch.set_alpha(0)
        self.ax.patch.set_alpha(0)
        self.ax.set_facecolor('none')
        self._update_pending = False
        self._last_histogram_params = (None, None, None, None, None, None, None, None)

    def resizeEvent(self, event):
        # Optionally, recalculate any subplot parameters here based on self.width() or self.height()
        # self.fig.subplots_adjust(left=0.16, right=0.95, bottom=0.3, top=0.9)
        # self.fig.tight_layout()  # Or set constrained_layout=True initially
        super().resizeEvent(event)

    def update_histogram(self, image, center, crop_size, sigma=None, intensity=None, background=None, peak=None, pointcolor="magenta"):
        """
        Instead of updating immediately, store the parameters and schedule an update
        only once per 16 ms (about 60 FPS).
        """

        self._last_histogram_params = (image, center, crop_size, sigma, intensity, background, peak, pointcolor)

        throttletime = 1

        if getattr(self.navigator, "looping", False):
            # fade the canvas back to half-opacity
            self.fig.patch.set_alpha(0.5)
            self.ax.clear()
            self.ax.axis("off")
            self.ax.text(
                0.5, 0.5, "Playback in progress",
                ha="center", va="center",
                transform=self.ax.transAxes,
                fontsize=14, color="grey"
            )
            self.draw()
            return
        
        if not self.isVisible():
            throttletime = 400
        # Save the latest parameters
        if not self._update_pending:
            self._update_pending = True
            # Schedule a delayed update
            QTimer.singleShot(throttletime, self._throttled_histogram_update)

    def _throttled_histogram_update(self):
        """This method is called after a short delay to perform the heavy update."""
        self._update_pending = False
        image, center, crop_size, sigma, intensity, background, peak, pointcolor = self._last_histogram_params
        self._do_update_histogram(image, center, crop_size, sigma, intensity, background, peak, pointcolor)

    def _do_update_histogram(self, image, center, crop_size, sigma=None, intensity=None, background=None, peak=None, pointcolor="magenta"):
        # --- Decide which portion of the image to use.
        if center is None or np.isnan(center[0]) or np.isnan(center[1]):
            subimage = image
            criteria_counts = None  # no criteria histogram
        else:
            H, W = image.shape
            half = crop_size // 2
            cx_int = int(round(center[0]))
            cy_int = int(round(center[1]))
            x1_crop = max(0, cx_int - half)
            x2_crop = min(W, cx_int + half)
            y1_crop = max(0, cy_int - half)
            y2_crop = min(H, cy_int + half)
            subimage = image[y1_crop:y2_crop, x1_crop:x2_crop]

        if subimage.size == 0:
            return

        # --- Compute histogram on the subimage.
        counts, bin_edges = np.histogram(subimage, bins=50)
        bin_centers = (bin_edges[:-1] + bin_edges[1:]) / 2

        # --- background
        self.current_background = background

        # --- Optionally compute a criteria histogram if sigma is provided.
        if (center is not None and not np.isnan(center[0]) and not np.isnan(center[1])
                and sigma is not None):
            sub_center_x = cx_int - x1_crop
            sub_center_y = cy_int - y1_crop
            y_indices, x_indices = np.indices(subimage.shape)
            dist = np.sqrt((x_indices - sub_center_x)**2 + (y_indices - sub_center_y)**2)
            sigma_mask = dist <= 2 * sigma
            criteria_pixels = subimage[sigma_mask]
            criteria_counts, _ = np.histogram(criteria_pixels, bins=bin_edges)
        else:
            criteria_counts = None

        # --- Clear axis, plot the histogram, overlay criteria if available.
        self.ax.clear()
        self.fig.patch.set_alpha(1.0)
        self.ax.set_facecolor("white")
        self.ax.set_xlabel("Pixel Intensity (A.U.)", fontsize=12)
        self.ax.set_ylabel("Count", fontsize=12)
        self.ax.tick_params(axis='both', which='major', labelsize=12)
        
        bar_width = bin_edges[1] - bin_edges[0]
        self.ax.bar(bin_centers, counts, width=bar_width, color='#7da1ff', edgecolor='black')
        if criteria_counts is not None:
            self.ax.bar(bin_centers, criteria_counts, width=bar_width, color=pointcolor, edgecolor='black')
        
        try:
            if (
                isinstance(background, (int, float)) and
                isinstance(peak, (int, float)) and
                math.isfinite(background) and
                math.isfinite(peak)
            ):
                actualpeak = peak+background
                # --- draw background threshold line only ---
                self.ax.axvline(
                    background,
                    color='black', linestyle='--', linewidth=1.5,
                    solid_capstyle='round',
                    dash_capstyle='round',
                    label=f"Bkgr: {background:.2f}"
                )
                self.ax.axvline(
                    actualpeak,
                    color=pointcolor, linestyle='--', linewidth=1.5,
                    solid_capstyle='round',
                    dash_capstyle='round',
                    label=f"Peak: {actualpeak:.2f}"
                )
                legend = self.ax.legend(
                    loc="upper right",
                    fontsize=10,
                    frameon=True,
                    labelspacing=0.5,
                    handlelength=2
                )
                # style the box
                frame = legend.get_frame()
                frame.set_facecolor("white")
                frame.set_alpha(0.8)
                frame.set_edgecolor("none")
                frame.set_boxstyle("round,pad=0.2")

            if not getattr(self.fig, "layout_engine", None):
                self.fig.subplots_adjust(left=0.17, right=0.95, bottom=0.3, top=0.9)
            self.draw()

        except np.linalg.LinAlgError:
            pass
            #print("Warning: Singular matrix encountered when drawing axvline")



from ._shared import *

class CustomSplitter(QSplitter):
    def __init__(self, orientation, parent=None, handle_y_offset_pct=0.5):
        super().__init__(orientation, parent)
        self._handle_y_offset_pct = handle_y_offset_pct

    def createHandle(self):
        return CustomSplitterHandle(self.orientation(), self, y_offset_pct=self._handle_y_offset_pct)


class CustomSplitterHandle(QSplitterHandle):
    def __init__(self, orientation, parent, y_offset_pct=0.5):
        super().__init__(orientation, parent)
        self._y_offset_pct = y_offset_pct  # store the percentage offset

    @property
    def y_offset_pct(self):
        return self._y_offset_pct

    @y_offset_pct.setter
    def y_offset_pct(self, value):
        # Optionally: validate that value is between 0 and 1
        if 0 <= value <= 1:
            self._y_offset_pct = value
            self.update()  # trigger repaint
        else:
            raise ValueError("y_offset_pct must be between 0 and 1.")

    def paintEvent(self, event):
        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing)
        pen = QPen(QColor("#888888"), 0)
        painter.setPen(pen)
        painter.setBrush(QColor("#888888"))
        if self.orientation() == Qt.Horizontal:
            handle_width = 2
            handle_length = 20
            x = int((self.width() - handle_width) / 2)
            available_height = self.height() - handle_length
            # Calculate y from the stored percentage:
            y = int(self._y_offset_pct * available_height)
            painter.drawRoundedRect(x, y, handle_width, handle_length, handle_width/2, handle_width/2)
        else:
            handle_width = 20
            handle_height = 2
            x = int((self.width() - handle_width) / 2)
            y = int((self.height() - handle_height) / 2)
            painter.drawRoundedRect(x, y, handle_width, handle_height, handle_height/2, handle_height/2)
        painter.end()


class RoundedFrame(QFrame):
    def __init__(self, parent=None, radius=10, bg_color="#FAFBFF", border_color=None):
        super().__init__(parent)
        self.radius = radius
        self.bg_color = QColor(bg_color)
        self.border_color = QColor(border_color) if border_color else QColor(0, 0, 0, 0)
        self.setAttribute(Qt.WA_OpaquePaintEvent)
        self.setAttribute(Qt.WA_NoSystemBackground)
        self.setAttribute(Qt.WA_TranslucentBackground)
        self.setFrameStyle(QFrame.NoFrame)

    def setBorderColor(self, color_str):
        self.border_color = QColor(color_str)
        self.update()

    def paintEvent(self, event):
        painter = QPainter(self)
        # Only draw in the widget’s rectangle.
        painter.setClipRect(self.rect())
        painter.setRenderHint(QPainter.Antialiasing)
        painter.setBrush(QBrush(self.bg_color))
        painter.setPen(QPen(self.border_color, 6))
        painter.drawRoundedRect(self.rect(), self.radius, self.radius)
        super().paintEvent(event)

